/* master.c
 * Neil Moore <neil@cs.uky.edu>
 * 22 May 2001
 *
 * Part of parstab.  See COPYRIGHT for copyright information.
 *
 */

/* XXX: should children seek work themselves (rather than going to the master)?
 *      it cuts comm down by one message, but increases the likelihood we will
 *      encounter a machine with no more work.
 */
#include <pvm3.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#define __USE_POSIX
#include <stdio.h>
#include <lp.h>
#include <lp_smrd.h>
#include <lp_util.h>
#include "misc.h"
#include "message.h"
#include "stable.h"
#include "timing.h"

#ifndef VERSION
#define VERSION "unknown"
#endif

/* The maximum length of a pathname */
#ifndef PATH_MAX
#define PATH_MAX 1024
#endif

/* Prototypes.  See below for the actual definitions. */
static void usage();
static void version(void);
static int to_int(const char *str);
static int send_program(struct lp *prog);
static int wait_loop(struct lp *prog);
static void kill_children(void);
static void print_model(struct lp *p, int *model);
static int do_waiting(machine_t m);
static int index_of(machine_t mac);
static int remove_machine(machine_t m);
static void handle_signal(int n);
/* Children is the array of slave tasks.  NTasks is the number of slave
 * tasks. */
static machine_t *Children=0;
static int NTasks = 3;

/* Should we log models? */
static int LogModels=1;

/* MaxModels is the maximum number of models to compute.  NModels is
 * the number of models we have so far. */
static int MaxModels = 0;
static int NModels = 0;

/* RunningJobs is the number of machines out there that have work.  When
 * it reaches zero, we have exhausted the search space.  HasWork[i] is 1
 * if Children[i] has work, 0 if it does not.  If a bug in the protocol
 * has given it work when it already had work, it will be set higher than
 * 1---this should be considered an error. */
static int RunningJobs=0;
static int *HasWork;

/* The file to log to.  Opened in append mode. */
FILE *logfile=0;

/* How verbosely should we log? */
int Logging=0;

/* Should we time verbosely? */
int Verbose_timing=0;


/* main.  See the text in usage() below for a description of the command-
 * line arguments */
int main(int argc, char *argv[]) 
{
	int i;

	char *progname;             /* name of this executable---the one to spawn */
	char *childargs[7];         /* arguments to slaves */
	char buff[20];              /* check frequency */
	char buff2[20];             /* balance depth */

	int result;                 /* Result code from PVM calls */
	int retval = 0;             /* Value to return with */

	char *runindir = NULL;      /* directory to run in */
	
	char *infilename = NULL;    /* input filename */
	FILE *infile = NULL;        /* input file stream */

	struct lp *prog;            /* the program */

	struct timeval start;
	struct timeval end;

	machine_t me;               /* my tid */
	machine_t parent;           /* my parent's tid, if I am a slave */

	progname = argv[0];

	me = pvm_mytid();
	
	if(me < 0)
		goto failure;	

	signal(SIGTERM, handle_signal);
	signal(SIGINT, handle_signal);
	signal(SIGSEGV, handle_signal);
	signal(SIGHUP, handle_signal);
	signal(SIGILL, handle_signal);
	signal(SIGABRT, handle_signal);
	signal(SIGQUIT, handle_signal);
	signal(SIGPIPE, handle_signal);

	parent = pvm_parent();
	if(parent < 0 && parent != PvmNoParent) {
		pvm_perror(progname);
		goto failure;
	}
	
	if(parent != PvmNoParent) {
		/* I am a slave */
		if(argc < 5 || argc > 6) {
			fprintf(stderr, "[%x]: ERROR: wrong number of arguments (%d)\n", me, argc);
			for(i=0; i<argc; ++i)
				fprintf(stderr, "[%x]: argv[%d] = \"%s\"", me, i, argv[i]);
		}

		if(strcmp(argv[1], VERSION)) {
			fprintf(stderr, "[%x]: ERROR: version mismatch!  I am `%s', master is `%s'",
				me, VERSION, argv[1]);
			goto failure;
		}
		
		if(argc == 7) {
			runindir = argv[6];
			if(chdir(runindir))
				fprintf(stderr, "[%x]: could not change to directory %s", me, runindir);
		}
		
		if(!strcmp(argv[2],"-LV"))
			Logging = 2;
		if(!strcmp(argv[2],"-L"))
			Logging = 1;
		if(!strcmp(argv[3],"-t")) {
			Timing = 1;
			Verbose_timing = 0;
			begin_time(TIM_MISC, "main");
		} else if(!strcmp(argv[3],"-tv")) {
			Timing = 1;
			Verbose_timing = 1;
			begin_time(TIM_MISC, "main");
		} else if(!strcmp(argv[3],"-ts")) {
			Timing = 1;
			TimeShort = 1;
			Verbose_timing = 0;
			begin_time(TIM_MISC, "main");
		}
		CheckFreq = to_int(argv[4]);
		BalanceDepth = to_int(argv[5]);
		
		logfile = fopen("parstab.log", "a");
			
		if(logfile == 0) {
			fprintf(stderr, "[%x]: could not open log file: %s", me, strerror(errno));
			Logging = 0;
		}

		for(i=0; i<argc; ++i)
			DO_LOG3(LP_LOG_INFO, "argv[%d] = \"%s\"", i, argv[i]);
		
		DO_LOG2(LP_LOG_DEBUG, "CheckFreq = %d", CheckFreq);
		DO_LOG2(LP_LOG_DEBUG, "BalanceDepth = %d", BalanceDepth);

		DO_LOG(LP_LOG_DEBUG, "calling pvm_notify()");
		if((result = pvm_notify(PvmTaskExit, MTAG_NOTIFY, 1, &parent))) {
			DO_LOG2(LP_LOG_DEBUG, "pvm_notify() failed (%d)", result);
		} else {
			DO_LOG(LP_LOG_DEBUG, "pvm_notify() succeeded");
		}
		
		/* find models, handle messages, etc.  do_stable waits for work if
		 * we don't have any. */
		result = do_stable();
		
		/* goto so the master code below won't be nested one more level */
		if(result)
			goto failure;
		goto done;
	}

	/* I am the master */
	TimingPrefix = "mpstim";
	gettimeofday(&start, NULL);
	/* parse the command line */
	while (*++argv != NULL) {
		if(!strcmp(*argv, "-d") || !strcmp(*argv, "--directory")) {
			++argv;
			if(*argv == NULL) {
				usage();
				return 1;
			}
			runindir = *argv;
		} else if(!strncmp(*argv, "-d", 2)) {
			runindir = *argv + 2;
		} else if(!strncmp(*argv, "--directory=", 12)) {
			runindir = *argv + 12;
		} else if(!strcmp(*argv, "-v") || !strcmp(*argv, "--version")) {
			version();
			return 0;
		} else if(!strcmp(*argv, "-h") || !strcmp(*argv, "--help")) {
			usage();
			return 0;
		} else if(!strcmp(*argv, "-1") || !strcmp(*argv, "--single")) {
			MaxModels = 1;
		} else if(!strcmp(*argv, "-a") || !strcmp(*argv, "--all")) {
			MaxModels = 0;
		} else if(!strcmp(*argv, "-c") || !strcmp(*argv, "--count")) {
			++argv;
			if(*argv == NULL || (MaxModels = to_int(*argv)) < 0) {
				usage();
				goto failure;
			}
		} else if(!strncmp(*argv, "-c", 2)) {
			if((MaxModels = to_int(*argv+2)) < 0) {
				usage();
				goto failure;
			}
		} else if(!strncmp(*argv, "--count=", 8)) {
			if((MaxModels = to_int(*argv+8)) < 0) {
				usage();
				goto failure;
			}
		} else if(!strcmp(*argv, "-f") || !strcmp(*argv, "--freq")) {
			++argv;
			if(*argv == NULL || (CheckFreq = to_int(*argv)) < 0) {
				usage();
				goto failure;
			}
		} else if(!strncmp(*argv, "-f", 2)) {
			if((CheckFreq = to_int(*argv+2)) < 0) {
				usage();
				goto failure;
			}
		} else if(!strncmp(*argv, "--freq=", 7)) {
			if((CheckFreq = to_int(*argv+7)) < 0) {
				usage();
				goto failure;
			}
		} else if(!strcmp(*argv, "-D") || !strcmp(*argv, "--depth")) {
			++argv;
			if(*argv == NULL || (BalanceDepth = to_int(*argv)) < 0) {
				usage();
				goto failure;
			}
		} else if(!strncmp(*argv, "-D", 2)) {
			if((BalanceDepth = to_int(*argv+2)) < 0) {
				usage();
				goto failure;
			}
		} else if(!strncmp(*argv, "--depth=", 8)) {
			if((BalanceDepth = to_int(*argv+8)) < 0) {
				usage();
				goto failure;
			}
		} else if(!strcmp(*argv, "-T") || !strcmp(*argv, "--tasks")) {
			++argv;
			if(*argv == NULL || (NTasks = to_int(*argv)) < 1) {
				usage();
				goto failure;
			}
		} else if(!strncmp(*argv, "-T", 2)) {
			if((NTasks = to_int(*argv+2)) < 1) {
				usage();
				goto failure;
			}
		} else if(!strncmp(*argv, "--tasks=", 8)) {
			if((NTasks = to_int(*argv+8)) < 1) {
				usage();
				goto failure;
			}
		} else if(!strcmp(*argv, "-l") || !strcmp(*argv, "--log")) {
			++argv;
			if(*argv == NULL || (Logging = to_int(*argv)) < 0 || Logging > LP_LOG_MAX ) {
				usage();
				goto failure;
			}
		} else if(!strncmp(*argv, "-l", 2)) {
			if((Logging = to_int(*argv+2)) < 0 || Logging > LP_LOG_MAX) {
				usage();
				goto failure;
			}
		} else if(!strncmp(*argv, "--log=", 6)) {
			if((Logging = to_int(*argv+8)) < 0 || Logging > LP_LOG_MAX) {
				usage();
				goto failure;
			}
		} else if(!strcmp(*argv, "-LM") || !strcmp(*argv, "--log-master")) {
			Logging = 3;
		} else if(!strcmp(*argv, "-l") || !strcmp(*argv, "--no-log")) {
			Logging = 0;
		} else if(!strcmp(*argv, "-tv") || !strcmp(*argv, "--time-verbose")) {
			Timing = 1;
			TimeShort = 0;
			Verbose_timing = 1;
		} else if(!strcmp(*argv, "-t") || !strcmp(*argv, "--time")) {
			Timing = 1;
			TimeShort = 0;
			Verbose_timing = 0;
		} else if(!strcmp(*argv, "-ts") || !strcmp(*argv, "--time-short")) {
			Timing = 1;
			TimeShort = 1;
			Verbose_timing = 0;
		} else if(!strcmp(*argv, "-n") || !strcmp(*argv, "--no-time")) {
			Timing = 0;
			Verbose_timing = 0;
		} else if(!strcmp(*argv, "-m") || !strcmp(*argv, "--models")) {
			LogModels = 1;
		} else if(!strcmp(*argv, "-q") || !strcmp(*argv, "--no-log-models")) {
			LogModels = 0;
		} else if(!strcmp(*argv, "-")) {
			infilename = NULL;
		} else if ((*argv)[0] == '-') {
			usage();
			goto failure;
		} else {
			infilename = *argv;
		}		
	}

	/* must open file before changing directories */
	if(infilename == NULL)
		infile = stdin;
	else
		infile = fopen(infilename, "r");
	
	if(runindir)
		if(chdir(runindir))
			fprintf(stderr, "could not change to directory %s", runindir);
	
	if(Timing)
		begin_time(TIM_MISC, "main");
	
	/* We still want logfile for models, even if Logging is 0. */
	logfile = fopen("parstab.log", "a");
	if(logfile == 0) {
		fprintf(stderr, "could not open log file: %s", strerror(errno));
		goto failure;
	}	

	
	prog = lp_malloc(sizeof(struct lp));

	if(infile == NULL) {
		DO_LOG3(LP_LOG_ERROR, "cannot open infile %s: %s", infilename,
			strerror(errno));
		goto failure;
	}

	DO_LOG2(LP_LOG_DEBUG, "CheckFreq = %d", CheckFreq);

	if(lp_sm_read_file(infile, prog) < 0) {
		const char *msg = lp_smrd_error();
		if(msg)
			DO_LOG2(LP_LOG_ERROR, "%s", msg);
		DO_LOG(LP_LOG_ERROR, "error in input; not running");
		goto failure;
	}	
	
	srand((unsigned) time(NULL));
	
	Children = lp_calloc(NTasks, sizeof(machine_t));
	HasWork = lp_calloc(NTasks, sizeof(int));

	childargs[0] = VERSION;
	switch(Logging) {
	    case 1:
		childargs[1] = "-L";
		break;
	    case 2:
		childargs[1] = "-LV";
		break;
	    case 0:
	    case 3:
	    default:
		childargs[1] = "-l";
		break;
	}
	childargs[2] = Timing ? (Verbose_timing ? "-tv" : (TimeShort ? "-ts" : "-t")) : "-n";
	sprintf(buff, "%d", CheckFreq);
	childargs[3] = buff;
	sprintf(buff2, "%d", BalanceDepth);
	childargs[4] = buff2;
	childargs[5] = runindir ? runindir : NULL;
	childargs[6] = NULL;

	begin_time(TIM_NET, "spawn");
	result = pvm_spawn(progname, childargs, PvmTaskDefault, NULL, NTasks, Children);
	end_time(TIM_NET, "spawn");
	
	if(result == 0) {
		DO_LOG2(LP_LOG_ERROR, "spawn failed (%d)", result);
		pvm_perror(progname);
		goto failure;
	}
	if(NTasks != result) {
		DO_LOG3(LP_LOG_ERROR, "spawned only %d of %d slaves", result, NTasks);
		NTasks = result;
	}
	printf("%x ", me);

	for(i=0; i<NTasks; ++i) {
		DO_LOG3(LP_LOG_INFO, "Child[%d] = %x", i, Children[i]);
		printf("%x ", Children[i]);
	}
	printf("\n");
	
	DO_LOG(LP_LOG_DEBUG, "calling pvm_notify()");
	if((result = pvm_notify(PvmTaskExit, MTAG_NOTIFY, NTasks, Children))) {
		DO_LOG2(LP_LOG_ERROR, "pvm_notify() failed (%d)", result);
	} else {
		DO_LOG(LP_LOG_DEBUG, "pvm_notify() succeeded");
	}
	
	if(send_program(prog)) {
		DO_LOG(LP_LOG_ERROR, "could not send program; killing children");
		kill_children();
		goto failure;
	}
	/* handle messages until we have enough models, we exhaust
	 * the search space, or all our children die. */
	if(wait_loop(prog)) {
		/* true = failure */
		DO_LOG(LP_LOG_ERROR, "wait_loop() failed; killing children");
		kill_children();
		goto failure;
	} else {
		/* false = success */
		double time;
		gettimeofday(&end, NULL);
		time = end.tv_sec - start.tv_sec;
		time += (end.tv_usec - start.tv_usec)*1.0E-6;
		printf("running time: %f sec\n", time);
		if(LogModels)
			DO_LOG2(LP_LOG_MODEL, "running time: %f sec\n", time);
		DO_LOG(LP_LOG_DEBUG, "wait_loop() finished successfully");
		kill_children();
	}
	
 done:
	retval = 0;
	
	if(0) { /* nasty trick */
 failure:
		if(retval == 0)
			retval = 1;
	}
	
	if(infile && infile != stdin)
		fclose(infile);
	if(logfile && logfile != stdout && logfile != stderr)
		fclose(logfile);
	pvm_exit();
	end_time(TIM_MISC, "main");
	stop_timing();
	return retval;
}

/* The main event loop for the master.  This loop waits for models from slaves,
 * prints models when it receives them, and forwards MSG_WAITING.  It also
 * keeps track of the number of machines with work---when this number drops to
 * zero, the search space has been exhausted.
 */
int wait_loop(struct lp *prog) 
{
	int result;
	struct message m;
	int idx;

	
	while(1) {
		if((result = wait_for_message(&m))) {
			DO_LOG2(LP_LOG_DEBUG, "wait_loop: error waiting for message (%d)", result);
			break;
		}
		switch(m.type) {
		    case MSG_WAITING:
			DO_LOG2(LP_LOG_INFO, "wait_loop: got MSG_WAITING from %x", m.from);
			idx = index_of(m.from);
			/* We can get MSG_WAITING from an already-waiting machine
			 * (if it got NAKWORK).  In such a case, don't lose.
			 */
			if(HasWork[idx]) {
				--HasWork[idx];
				DO_LOG3(LP_LOG_DEBUG, "wait_loop: HasWork(%x) = %d", m.from, HasWork[idx]);
			
				/* XXX: be careful.  Probably a race condition here */
				if(--RunningJobs == 0) {
					DO_LOG(LP_LOG_INFO, "wait_loop: RunningJobs = 0");
					return 0;
				}
				DO_LOG2(LP_LOG_DEBUG, "wait_loop: RunningJobs = %d", RunningJobs);
			}
			
			if((result = do_waiting(m.from))) {
				DO_LOG2(LP_LOG_ERROR, "wait_loop: do_waiting() failed (%d)", result);
			} else {
				DO_LOG(LP_LOG_DEBUG, "wait_loop: do_waiting() succeeded");
			}
			break;

		    case MSG_MODEL:
		    case MSG_MODEL_UPD:
			DO_LOG2(LP_LOG_INFO, "wait_loop: got MSG_MODEL from %x", m.from);
			
			if(m.size != prog->natoms) {
				DO_LOG4(LP_LOG_ERROR, "wait_loop: ERROR: bad model size from %x (its %d != my %d)",
					m.from, m.size, prog->natoms);
				free(m.data);
			} else {
				++NModels;
				print_model(prog, m.data);
				free(m.data);
				if(NModels == MaxModels) {
					DO_LOG2(LP_LOG_INFO, "wait_loop: found all %d models", MaxModels);
					return 0;
				}
			}
			break;
			
		    case MSG_SENTJOB:
			DO_LOG3(LP_LOG_INFO, "wait_loop: got MSG_SENTJOB from %x for %x", m.from, m.misc);
			++HasWork[index_of(m.misc)];
			DO_LOG3(LP_LOG_DEBUG, "wait_loop: HasWork(%x) = %d", m.misc, HasWork[index_of(m.misc)]);
			++RunningJobs;
			DO_LOG2(LP_LOG_DEBUG, "wait_loop: RunningJobs = %d", RunningJobs);
			break;

		    case MSG_DIED:
			DO_LOG2(LP_LOG_ERROR, "wait_loop: machine %x died", m.from);
			if(remove_machine(m.from))
				return -1;
			break;
		    default:
			not_expecting(m.type);
		}
	}
	return 1;
}


int send_program(struct lp *prog)
{
	int result;
	int i;

	for(i=0; i<NTasks; ++i) {
		DO_LOG3(LP_LOG_INFO, "send_program: sending to child %d (%x)", i, Children[i]);
		if((result = send_prog(Children[i], prog, i, NTasks))) {
			DO_LOG4(LP_LOG_ERROR, "send_program: sending to child %d (%x) failed (%d)",
				i, Children[i], result);
			if(remove_machine(Children[i]))
				return -1;
			--i;
			continue;
		}
		DO_LOG3(LP_LOG_INFO, "send_program: sent to child %d (%x)",
			i, Children[i]);
		
		++HasWork[i];
		DO_LOG3(LP_LOG_INFO, "send_program: HasWork(%x) = %d",
			Children[i], HasWork[i]);
		++RunningJobs;
		DO_LOG2(LP_LOG_INFO, "send_program: RunningJobs = %d",
			RunningJobs);
	}

	return 0;
}


/* Like atoi(), but returns a negative number if the string is non-numeric */
int to_int(const char *str)
{
	long result;
	char *end=NULL;
	if(str == NULL)
		return 0;
	
	result = strtol(str, &end, 0);

	if(*str != '\0' && *end == '\0')
		return (int)result;
	else
		return -1;
}

void kill_children(void) 
{
	int i;
	
	for(i=0; i<NTasks; ++i) {
		DO_LOG3(LP_LOG_INFO, "asking child %d (%x) to die", i, Children[i]);
		send_die(Children[i]);
	}
	if(NTasks == 0)
		return;
	DO_LOG(LP_LOG_INFO, "waiting for children to die");
	
	begin_time(TIM_WAIT, "sleep_before_kill");
	sleep(3);
	end_time(TIM_WAIT, "sleep_before_kill");
	begin_time(TIM_NET, "pvm_kill");
	for(i=0; i<NTasks; ++i) {
		DO_LOG3(LP_LOG_INFO, "killing child %d (%x)", i, Children[i]);
		pvm_kill(Children[i]);
	}
	end_time(TIM_NET, "pvm_kill");
}


void version(void) 
{
	printf("parstab %s, from lpsms %s, using liblp %s\n", VERSION, lp_sms_version(), lp_version());
}

void usage(void) 
{
	printf("usage: parstab [options] [filename]\n");
	printf("Compute the stable models of a logic program.\n\n");

	printf(" -T, --tasks=N          spawn N processes (default 3)\n");
	printf(" -1, --single           find only the first model\n");
	printf(" -a, --all              find all stable models (default)\n");
	printf(" -c, --count=N          find the first N stable models\n");
	printf(" -f, --freq=N           check for machines wanting work every N forced atoms;\n");
	printf("                         0 == choice points only (default 100).\n");
	printf(" -D, --depth=N          go N choicepoints deep to find work to transfer;\n");
	printf("                         1 == first CP, 0 == as deep as possible (default 1).\n");
	printf(" -LM, --log-master      log at the master only.\n");
	printf(" -l, --log=N            log level.  0=silent, 1=errors only, 2=info, 3=debug\n");
	printf("                         (default 1).  Logging of models is controlled by -m/-q\n");
	printf(" -m, --log-models       log models to parstab.log (default)\n");
	printf(" -q, --no-log-models    compute models but do not print them\n");
	printf(" -t, --time             perform timing to pstim-[tid].log\n");
	printf(" -tv, --time-verbose    perform very verbose timing to pstim-[tid].log\n");
	printf(" -ts, --time-short      perform short timing to pstim-[tid].log\n");
	printf(" -n, --no-time          do not perform timing (default)\n");
	printf(" -d, --directory=dir    perform logging and timing in dir (no default)\n");
	printf(" -v, --version          print version information to stdout\n");
	printf(" -h, --help             display this message\n\n");

	printf("Input is read from filename (defaults to stdin).\n");	
}

void print_model(struct lp *p, int *model) 
{
	int i;
	
	if(LogModels) {
		begin_time(TIM_LOG, "print_model");
		fprintf(logfile, "[%x] Model %d: ", pvm_mytid(), NModels);
		for(i=0; i<p->natoms; i++) {
			if(model[i] == LP_L_TRUE)
				fprintf(logfile, "%s ", p->atoms[i].name);
			else if(model[i] == LP_L_UNKNOWN)
				fprintf(logfile, "ERROR(%s UNKNOWN) ",
					p->atoms[i].name);
		}
		fprintf(logfile,"\n");
		fflush(logfile);
		end_time(TIM_LOG, "print_model");
	}
}

int do_waiting(machine_t machine) 
{
	int result;
	machine_t send_to;
	int i;
	int lb;

	begin_time(TIM_COMPUTE, "find_rand_machine");
	
	lb = (int)(((double)NTasks * rand())/(RAND_MAX+1.0));
	i=lb;
	
	DO_LOG3(LP_LOG_DEBUG, "do_waiting: trying (random) %d [%x]", i, Children[i]);

	do {
		++i;
		if(i == NTasks)
			i = 0;
		if(Children[i] == machine)
			continue;
		if(HasWork[i])
			break;
	} while(i != lb);

	end_time(TIM_COMPUTE, "find_rand_machine");
	
	DO_LOG3(LP_LOG_DEBUG, "do_waiting: actually using %d [%x]", i, Children[i]);
	send_to = Children[i];

	/* can't use send_waiting because we need to fake the from field */
	begin_time(TIM_NET, "resend_waiting");
	pvm_initsend(PvmDataDefault);
	if((result = pvm_packf("%d%d", MSG_WAITING, machine))) {
		DO_LOG(LP_LOG_ERROR, "do_waiting: error packing header");
		end_time(TIM_NET, "resend_waiting_err");
		return result;
	}
	
	DO_LOG3(LP_LOG_INFO, "do_waiting: sending MSG_WAITING to %x from %x", send_to, machine);
	if((result = pvm_send(send_to, MTAG_MSG))) {
		DO_LOG2(LP_LOG_ERROR, "do_waiting: error sending MSG_WAITING (%d)", result);
		end_time(TIM_NET, "resend_waiting_err");
		return result;
	}
	DO_LOG(LP_LOG_DEBUG, "do_waiting: sent MSG_WAITING");
	end_time(TIM_NET, "resend_waiting");
	return 0;
}

int index_of(machine_t mac) 
{
	int i;
	begin_time(TIM_COMPUTE, "index_of");
	for(i=0; i<NTasks; ++i)
		if(Children[i] == mac) {
			end_time(TIM_COMPUTE, "index_of");
			return i;
		}
	end_time(TIM_COMPUTE, "index_of_not_found");
	return -1;
}

int remove_machine(machine_t m) 
{
	int idx,i;	

	if(NTasks == 0)
		return 1;
	
	idx = index_of(m);
	if(idx == -1) {
		DO_LOG2(LP_LOG_ERROR, "remove_machine: I see no machine %x here", m);
		return 0;
	}
	RunningJobs -= HasWork[idx];
	begin_time(TIM_COMPUTE, "remove_machine_adjust");
	DO_LOG3(LP_LOG_INFO, "remove_machine: machine %x (Children[%d]) exited", m, idx);
	--NTasks;
	for(i=idx; i<NTasks; ++i) {
		Children[i] = Children[i+1];
		HasWork[i] = HasWork[i+1];
	}
	
	DO_LOG2(LP_LOG_DEBUG, "remove_machine: adjusted children---%d left", NTasks);
	end_time(TIM_COMPUTE, "remove_machine_adjust");
	
	if(NTasks == 0)
		return 1;
	
	return 0;
}

void handle_signal(int n)
{
	if(logfile && Logging >= LP_LOG_ERROR)
		DO_LOG2(LP_LOG_ERROR, "received signal %d---exiting", n);
	exit(1);
} 
