/* message.c
 * Neil Moore <neil@cs.uky.edu>
 * 22 May 2001
 *
 * Part of parstab.  See COPYRIGHT for copyright information.
 */

#include <pvm3.h>
#include <stdlib.h>
#include <stdio.h>
#include <lp.h>
#include <lp_util.h>
#include "message.h"
#include "misc.h"

static int read_msg_prog(struct lp *prog);

char *MessageNames[] = {
	"MSG_NULL", "MSG_STACK", "MSG_DIE", "MSG_WAITING",
	"MSG_MODEL", "MSG_NAKWORK", "MSG_SENTJOB",
	"MSG_DIED", "MSG_PROG", "MSG_STACK_UPD", "MSG_MODEL_UPD", 0
};

int wait_for_message(struct message *msg) 
{
	int result=0;
	int junk1, tag=0, junk2;

	begin_time(TIM_NET, "wait_for_message");
	begin_time(TIM_WAIT, "wait_for_message_recv");
	result = pvm_recv(-1, -1);
	end_time(TIM_WAIT, "wait_for_message_recv");
	if(result < 0) {
		DO_LOG2(LP_LOG_ERROR, "wait_for_message: recv failed (%d)", result);
		goto done;
	}

	pvm_bufinfo(result, &junk1, &tag, &junk2);
	
	result = 0;
	
	if(tag == MTAG_MSG) 
		result=parse_message(msg);
	else if(tag == MTAG_NOTIFY)
		result=notify_message(msg);
	else {
		DO_LOG2(LP_LOG_ERROR, "wait_for_message: bad tag (%d)", tag);
		result=-1;
	}
 done:
	end_time(TIM_NET, result ? "wait_for_message_err" : "wait_for_message");
	return result;
}

int notify_message(struct message *msg) 
{
	int result=0;
	char pbuf[80];

	begin_time(TIM_NET, "notify_message");
	msg->type = MSG_DIED;
	if((result = pvm_upkint(&(msg->from), 1, 1)))
		DO_LOG2(LP_LOG_ERROR, "notify_message: error unpacking type (%d)", result);
	else	
		sprintf(pbuf, "notify_message about %x", msg->from);

	end_time(TIM_NET, result ? "notify_message_err" : pbuf);
	return result;
}


int parse_message(struct message *msg) 
{
	int result=0;
	char pbuf[80];
	int smallp, i;
	unsigned int *dbuff;

	begin_time(TIM_NET, "parse_message");
	
	if((result = pvm_upkint(&(msg->type), 1, 1))) {
		DO_LOG2(LP_LOG_ERROR, "parse_message: error unpacking type (%d)", result);
		goto done;
	}
	
	
	if((result = pvm_upkint(&(msg->from), 1, 1))) {
		DO_LOG2(LP_LOG_ERROR, "parse_message: error unpacking from (%d)", result);
		goto done;
	}

	switch(msg->type) {
	    case MSG_MODEL:
		if((result = pvm_upkint(&(msg->size), 1, 1))) {
			DO_LOG2(LP_LOG_ERROR, "parse_message: error unpacking size (%d)", result);
			goto done;
		}
		msg->data = lp_malloc(sizeof(int) * msg->size);
		if((result = pvm_upkint(msg->data, msg->size, 1))) {
			DO_LOG2(LP_LOG_ERROR, "parse_message: error unpacking model (%d)", result);
			goto done;
		}
		break;
	    case MSG_MODEL_UPD:
		if((result = pvm_upkint(&(msg->size), 1, 1))) {
			DO_LOG2(LP_LOG_ERROR, "parse_message: error unpacking size (%d)", result);
			goto done;
		}
		dbuff = lp_malloc(sizeof(unsigned int) * (msg->size + 31)/32);
		msg->data = lp_malloc(sizeof(int) * msg->size);
		if((result = pvm_upkuint(dbuff, (msg->size+31)/32, 1))) {
			DO_LOG2(LP_LOG_ERROR, "parse_message: error unpacking model (%d)", result);
			goto done;
		}
		for(i=0; i<msg->size; ++i)
			((int *)(msg->data))[i] = (dbuff[i/32] & (1U << i%32)) ? LP_L_TRUE : LP_L_FALSE;
		free(dbuff);
		break;
	    case MSG_STACK:
		if((result = pvm_upkint(&(msg->size), 1, 1))) {
			DO_LOG2(LP_LOG_ERROR, "parse_message: error unpacking size (%d)", result);
			goto done;
		}
		msg->data = lp_malloc(sizeof(int) * (2*msg->size));
		if((result = pvm_upkint(msg->data, msg->size, 1))) {
			DO_LOG2(LP_LOG_ERROR, "parse_message: error unpacking stack (%d)", result);
			goto done;
		}
		if((result = pvm_upkint(((int *)msg->data) + msg->size, msg->size, 1))) {
			DO_LOG2(LP_LOG_ERROR, "parse_message: error unpacking aux (%d)", result);
			goto done;
		}
		break;
	    case MSG_STACK_UPD:
		if((result = pvm_upkint(&(msg->size), 1, 1))) {
			DO_LOG2(LP_LOG_ERROR, "parse_message: error unpacking size (%d)", result);
			goto done;
		}
		if((result = pvm_upkint(&smallp, 1, 1))) {
			DO_LOG2(LP_LOG_ERROR, "parse_message: error unpacking small flag (%d)", result);
			goto done;
		}
		dbuff = lp_malloc(sizeof(int) * (2*msg->size));
		msg->data = dbuff;
		
		if(smallp) {
			unsigned short *thing = lp_malloc(sizeof(unsigned short) * (2*msg->size));
			
			if((result = pvm_upkushort(thing, msg->size, 1))) {
				DO_LOG2(LP_LOG_ERROR, "parse_message: error unpacking stack (%d)", result);
				goto done;
			}
			for(i=0; i < msg->size; ++i) {
				dbuff[i] = thing[i] & 0x3fff;
				dbuff[msg->size+i] = ((thing[i] & 0x8000) >> 15);
			}
			free(thing);
		} else {
			
			if((result = pvm_upkint(msg->data, msg->size, 1))) {
				DO_LOG2(LP_LOG_ERROR, "parse_message: error unpacking stack (%d)", result);
				goto done;
			}
			for(i=0; i<msg->size; ++i) {
				dbuff[msg->size + i] = ((unsigned int)(dbuff[i] & 0x80000000) >> 31);
				dbuff[i] &= 0x3fffffff;
			}
		}
		break;
	    case MSG_PROG:
		if((result = pvm_upkint(&(msg->misc), 1, 1))) {
			DO_LOG2(LP_LOG_ERROR, "parse_message: error unpacking child number (%d)", result);
			goto done;
		}
		if((result = pvm_upkint(&(msg->misc2), 1, 1))) {
			DO_LOG2(LP_LOG_ERROR, "parse_message: error unpacking number of children (%d)", result);
			goto done;
		}
		msg->data = lp_malloc(sizeof(struct lp));
		if((result = read_msg_prog((struct lp *)msg->data)))
			goto done;
		break;
	    case MSG_SENTJOB:
		if((result = pvm_upkint(&(msg->misc), 1, 1))) {
			DO_LOG2(LP_LOG_ERROR, "parse_message: error unpacking options (%d)", result);
			goto done;
		}
		break;
	    case MSG_WAITING:
	    case MSG_DIE:
	    case MSG_DIED:
	    case MSG_NAKWORK:
		break;
	    default:
		DO_LOG2(LP_LOG_ERROR, "parse_message: bad message (%d)", msg->type);
		result = -1;
		goto done;
	}
	
 done:
	if(result==0)
		sprintf(pbuf, "parse_message: %s from %x", message_name(msg->type), msg->from);
	end_time(TIM_NET, result ? "parse_message_err" : pbuf);
	return result;
}


int send_generic(machine_t m, int msgtype) 
{
	int result=0;
	char pbuf[80];

	sprintf(pbuf, "send_generic (%s) to %x", message_name(msgtype), m);
	begin_time(TIM_NET, pbuf);
	pvm_initsend(PvmDataDefault);
	if((result = pvm_packf("%d%d", msgtype, pvm_mytid())))
		DO_LOG3(LP_LOG_ERROR, "send_generic(%s): error packing header (%d)", message_name(msgtype), result);
	else if((result = pvm_send(m, MTAG_MSG)))
		DO_LOG4(LP_LOG_ERROR, "send_generic(%s): error sending to %x (%d)", message_name(msgtype), m, result);

	end_time(TIM_NET, result ? "send_generic_err" : "send_generic");
	return result;
}



int send_nakwork(machine_t m)
{
	return send_generic(m, MSG_NAKWORK);
}


int send_died(machine_t m)
{
	return send_generic(m, MSG_DIED);
}


int send_die(machine_t m)
{
	return send_generic(m, MSG_DIE);
}


int send_waiting(machine_t m)
{
	return send_generic(m, MSG_WAITING);
}


int send_sentjob(machine_t machine, machine_t to_whom)
{
	int result = 0;
	char pbuf[64];

	sprintf(pbuf, "send_sentjob to %x", machine);
	begin_time(TIM_NET, pbuf);
	pvm_initsend(PvmDataDefault);
	if((result = pvm_packf("%d%d%d", MSG_SENTJOB, pvm_mytid(), to_whom)))
		DO_LOG2(LP_LOG_ERROR, "send_sentjob: error packing header (%d)", result);
	else if((result = pvm_send(machine, MTAG_MSG)))
		DO_LOG3(LP_LOG_ERROR, "send_sentjob: error sending to %x (%d)", machine, result);

	end_time(TIM_NET, result ? "send_sentjob_err" : "send_sentjob");
	
	return result;
}


int send_model(machine_t machine, struct lp *prog, int *model) 
{
	int result=0;
	
	begin_time(TIM_NET, "send_model");

	pvm_initsend(PvmDataDefault);
	
	if((result = pvm_packf("%d%d%d", MSG_MODEL, pvm_mytid(), prog->natoms)))
		DO_LOG2(LP_LOG_ERROR, "send_model: error packing header (%d)", result);
	else if((result = pvm_packf("%*d", prog->natoms, model)))
		DO_LOG2(LP_LOG_ERROR, "send_model: error packing atoms (%d)", result);
	else if((result = pvm_send(machine, MTAG_MSG)))
		DO_LOG3(LP_LOG_ERROR, "send_model: error sending to %x (%d)", machine, result);
	
	end_time(TIM_NET, result ? "send_model_err" : "send_model");
	return result;
}


int send_model_upd(machine_t machine, struct lp *prog, int *model) 
{
	int result=0;
	unsigned int *ucopy;
	int i;
	
		
	begin_time(TIM_NET, "send_model_upd");

	ucopy = lp_calloc((prog->natoms + 31)/32, sizeof(unsigned int));
	for(i=0; i<prog->natoms; i++) {
		if(model[i] == LP_L_TRUE)
			ucopy[i/32] |= (1U << i%32);
		if(model[i] == LP_L_UNKNOWN)
			DO_LOG3(LP_LOG_ERROR, "send_model_upd: ERROR: missing atom %d (%s)", i, prog->atoms[i].name);
	}

	pvm_initsend(PvmDataDefault);
	
	if((result = pvm_packf("%d%d%d", MSG_MODEL_UPD, pvm_mytid(), prog->natoms)))
		DO_LOG2(LP_LOG_ERROR, "send_model: error packing header (%d)", result);
	else if((result = pvm_pkuint(ucopy, prog->natoms, 1)))
		DO_LOG2(LP_LOG_ERROR, "send_model: error packing atoms (%d)", result);
	else if((result = pvm_send(machine, MTAG_MSG)))
		DO_LOG3(LP_LOG_ERROR, "send_model: error sending to %x (%d)", machine, result);

	free(ucopy);
	end_time(TIM_NET, result ? "send_model_upd_err" : "send_model_upd");
	return result;
}


int send_stack_upd(machine_t machine, int top, int *stack, int *aux) 
{
	int result=0;
	char pbuf[64];
	int i;
	int smallp;
	unsigned int *copy = NULL;
	unsigned short *hcopy = NULL;

	sprintf(pbuf, "send_stack_upd to %x", machine);
	begin_time(TIM_NET, pbuf);
	
	pvm_initsend(PvmDataDefault);
		smallp = (top < 0x8000);
	
	if(smallp)
		hcopy = lp_calloc(top, sizeof(unsigned short));
	else
		copy = lp_calloc(top, sizeof(unsigned int));

	for(i=0;i<top;++i) {
		if(smallp)
			hcopy[i] = (unsigned short)stack[i] | ((unsigned short)aux[i] << 15);
		else
			hcopy[i] = (unsigned int)stack[i] | ((unsigned int)aux[i] << 31);
	}
	
	
	if((result = pvm_packf("%d%d%d%d", MSG_STACK_UPD, pvm_mytid(), top, smallp)))
		DO_LOG2(LP_LOG_ERROR, "send_stack_upd: error packing header (%d)", result);
	else if(smallp && (result = pvm_pkushort(hcopy, top, 1)))
		DO_LOG2(LP_LOG_ERROR, "send_stack_upd: error packing stack (small) (%d)", result);
	else if(!smallp && (result = pvm_pkuint(copy, top, 1)))
		DO_LOG2(LP_LOG_ERROR, "send_stack_upd: error packing stack (%d)", result);
	else if((result = pvm_send(machine, MTAG_MSG)))
		DO_LOG3(LP_LOG_ERROR, "send_stack_upd: error sending to %x (%d)", machine, result);

	free(smallp ? (void *)hcopy : (void *)copy);
	end_time(TIM_NET, result ? "send_stack_upd_err" : pbuf);
	return result;
}

int send_stack(machine_t machine, int top, int *stack, int *aux)
{
	int result=0;
	char pbuf[64];
	int *auxcp;
	int i;

	sprintf(pbuf, "send_stack to %x", machine);
	begin_time(TIM_NET, pbuf);

	auxcp = lp_malloc(sizeof(int)*top);
	for(i=0; i<top; ++i)
		auxcp[i] = aux[i] & 0x1;
	
	pvm_initsend(PvmDataDefault);

	if((result = pvm_packf("%d%d%d", MSG_STACK, pvm_mytid(), top)))
		DO_LOG2(LP_LOG_ERROR, "send_stack: error packing header (%d)", result);
	else if((result = pvm_packf("%*d%*d", top, stack, top, auxcp)))
		DO_LOG2(LP_LOG_ERROR, "send_stack: error packing stack (%d)", result);
	else if((result = pvm_send(machine, MTAG_MSG)))
		DO_LOG3(LP_LOG_ERROR, "send_stack: error sending to %x (%d)", machine, result);

	free(auxcp);
	end_time(TIM_NET, result ? "send_stack_err" : pbuf);
	return result;
}


int send_prog(machine_t machine, struct lp *prog, int cldnum, int children) 
{
	int result=0;
	int i,k;
	int *pos=0, *neg=0;
	int dummy = 0;
	char pbuf[96];

	sprintf(pbuf, "send_prog to %x (%d/%d)", machine, cldnum, children);
	begin_time(TIM_NET, pbuf);
	
	pvm_initsend(PvmDataDefault);
	if((result = pvm_packf("%d%d%d%d%d%d%d", MSG_PROG, pvm_mytid(), cldnum,
			       children, prog->nrules, prog->natoms, prog->size))) {
		DO_LOG2(LP_LOG_ERROR, "send_prog: error packing header (%d)", result);
		goto done;
	}
	/* The slaves don't need the atom names, so don't include them */

	if(prog->nposreq) {
		pos = lp_malloc(prog->nposreq * sizeof(int));
		for(i=0; i<prog->nposreq; ++i) {
			pos[i] = lp_apos(prog,prog->posreq[i]);
		}
	}

	if(prog->nnegreq) {
		neg = lp_malloc(prog->nnegreq * sizeof(int));
		for(i=0; i<prog->nnegreq; ++i) {
			neg[i] = lp_apos(prog,prog->negreq[i]);
		}
	}

	/* Avoid touching a null pointer, even if we access zero elements of it */
	if((result = pvm_packf("%d%d%*d%*d", prog->nposreq, prog->nnegreq, prog->nposreq,
			       pos ? pos : &dummy, prog->nnegreq, neg ? neg : &dummy))) {
		DO_LOG2(LP_LOG_ERROR, "send_prog: error packing requirements (%d)", result);
		free(pos);
		free(neg);
		goto done;
	}

	if(pos)
		free(pos);
	if(neg)
		free(neg);

	for(i=0; i < prog->nrules; ++i) {
		struct lp_rule *j = &(prog->rules[i]);

		pos = neg = 0;

		if(j->npos) {
			pos = lp_malloc(j->npos * sizeof(int));
			for(k=0; k<j->npos; ++k) {
				pos[k] = lp_apos(prog,j->pos[k]);
			}
		}

		if(j->nneg) {
			neg = lp_malloc(j->nneg * sizeof(int));
			for(k=0; k<j->nneg; ++k) {
				neg[k] = lp_apos(prog,j->neg[k]);
			}
		}
		
		if((result = pvm_packf("%d%d%d%*d%*d", lp_apos(prog,j->head),
				                       j->npos, j->nneg, 
				                       j->npos, pos ? pos : &dummy,
				                       j->nneg, neg ? neg : &dummy))) {
			DO_LOG3(LP_LOG_ERROR, "send_prog: error packing rule %d (%d)", i, result);
			if(pos)
				free(pos);
			if(neg)
				free(neg);
			goto done;
		}

		if(pos)
			free(pos);
		if(neg)
			free(neg);
	}
	
	if((result = pvm_send(machine, MTAG_MSG)))
		DO_LOG3(LP_LOG_ERROR, "send_prog: error sending to %x (%d)", machine, result);
 done:
	end_time(TIM_NET, result ? "send_prog_err" : "send_prog");
	return result;
}

void not_expecting(int type) 
{
	DO_LOG3(LP_LOG_ERROR, "bad message sequence: not expecting %s (%d)\n",
		message_name(type), type);
}

int read_msg_prog(struct lp *prog) 
{
	int result=0, i, j;
	int natoms, nrules, size;
	int *pos, *neg;

	begin_time(TIM_NET, "read_msg_prog");
	
	if((result = pvm_upkint(&nrules, 1, 1))) {
		DO_LOG2(LP_LOG_ERROR, "read_msg_prog: error unpacking nrules (%d)", result);
		goto done;
	}
	
	if((result = pvm_upkint(&natoms, 1, 1))) {
		DO_LOG2(LP_LOG_ERROR, "read_msg_prog: error unpacking natoms (%d)", result);
		goto done;
	}

	if((result = pvm_upkint(&size, 1, 1))) {
		DO_LOG2(LP_LOG_ERROR, "read_msg_prog: error unpacking size (%d)", result);
		goto done;
	}

	lp_init_lp(prog, nrules, natoms, size);
	
	for(i=0; i<natoms; ++i) {
		sprintf(prog->atoms[i].name, "a%d", i);
	}

	if((result = pvm_upkint(&(prog->nposreq), 1, 1))) {
		DO_LOG2(LP_LOG_ERROR, "read_msg_prog: error unpacking nposreq (%d)", result);
		goto done;
	}
	if((result = pvm_upkint(&(prog->nnegreq), 1, 1))) {
		DO_LOG2(LP_LOG_ERROR, "read_msg_prog: error unpacking nnegreq (%d)", result);
		goto done;
	}
	
	pos = lp_malloc(prog->nposreq * sizeof(int));
	prog->posreq = lp_malloc(prog->nposreq * sizeof(struct lp_atom *));
	if((result = pvm_upkint(pos, prog->nposreq, 1))) {
		DO_LOG2(LP_LOG_ERROR, "read_msg_prog: error unpacking posreq (%d)", result);
		goto done;
	}
	for(i=0; i<prog->nposreq; ++i)
		prog->posreq[i] = &(prog->atoms[pos[i]]);
	free(pos);

	neg = lp_malloc(prog->nnegreq * sizeof(int));
	prog->negreq = lp_malloc(prog->nnegreq * sizeof(struct lp_atom *));
	if((result = pvm_upkint(neg, prog->nnegreq, 1))) {
		DO_LOG2(LP_LOG_ERROR, "read_msg_prog: error unpacking negreq (%d)", result);
		goto done;
	}
	for(i=0; i<prog->nnegreq; ++i)
		prog->negreq[i] = &(prog->atoms[neg[i]]);
	free(neg);
	
	
	for(i=0; i<nrules; ++i) {
		int head;
		struct lp_rule *r = &(prog->rules[i]);

		if((result = pvm_upkint(&head, 1, 1))) {
			DO_LOG3(LP_LOG_ERROR, "read_msg_prog: error unpacking head for rule %d (%d)", i, result);
			goto done;
		}
		r->head = &(prog->atoms[head]);
		
		if((result = pvm_upkint(&(r->npos), 1, 1))) {
			DO_LOG3(LP_LOG_ERROR, "read_msg_prog: error unpacking npos for rule %d (%d)", i, result);
			goto done;
		}
		if((result = pvm_upkint(&(r->nneg), 1, 1))) {
			DO_LOG3(LP_LOG_ERROR, "read_msg_prog: error unpacking nneg for rule %d (%d)", i, result);
			goto done;
		}

		pos = lp_malloc(r->npos * sizeof(int));
		if((result = pvm_upkint(pos, r->npos, 1))) {
			DO_LOG3(LP_LOG_ERROR, "read_msg_prog: error unpacking pos for rule %d (%d)", i, result);
			goto done;
		}
		for(j=0; j < r->npos; ++j)
			r->pos[j] = &(prog->atoms[pos[j]]);
		free(pos);

		neg = lp_malloc(r->nneg * sizeof(int));
		if((result = pvm_upkint(neg, r->nneg, 1))) {
			DO_LOG3(LP_LOG_ERROR, "read_msg_prog: error unpacking neg for rule %d (%d)", i, result);
			goto done;
		}
		for(j=0; j < r->nneg; ++j)
			r->neg[j] = &(prog->atoms[neg[j]]);
		free(neg);
	}

 done:
	end_time(TIM_NET, result ? "read_msg_prog_err" : "read_msg_prog");
	return result;
}


char *message_name(int type) {
	if(type >= 0 || type <= MSG_MAX)
		return MessageNames[type];
	else
		return "(UNKNOWN)";
}
