/* stable.c
 * Neil Moore <neil@cs.uky.edu>
 * 22 May 2001
 *
 * Part of parstab.  See COPYRIGHT for copyright information.
 */

/*
 * TODO: make it faster.  Even with similar choice point statistics, we
 *   are often much slower than smodels.  On the other hand, we are often
 *   much faster.
 *
 * TODO: keep track of how deep each tree is (in master).
 *
 * TODO: keep track of how many CPs we have; if it drops to zero, don't
 *       bother probing for messages in the loop (-f option).
 */

#include <string.h>
#include <stddef.h>
#include <lp.h>
#include <unistd.h>
#include <lp_util.h>
#include <lp_fks.h>
#include <lp_lm.h>
#include <lp_wfs.h>
#include <lp_queue.h>
#include <pvm3.h>

#include "timing.h"
#include "message.h"
#include "stable.h"
#include "misc.h"

#define TRY_SINGLE_INIT /* */
#define USE_SHORTER_MSGS /* */

#ifdef USE_SHORTER_MSGS
# define send_stack_CH send_stack_upd
# define send_model_CH send_model_upd
#else
# define send_stack_CH send_stack
# define send_model_CH send_model
#endif

#define DO_VERBOSE(X) do { if(Logging > 1) { X; } } while(0)

int CheckFreq=100;  /* Number of forced atoms before checking for messages */
int BalanceDepth=1;  /* Depth (# of CPs) to go in tree for work.  1 = first CP, etc; 0 = as deep as possible */

static struct mac_queue *WaitingMachines;

#define AUX_EXP_NEG 0
#define AUX_EXP_POS 1
#define AUX_NEG     2
#define AUX_POS     3

static const char *AuxNames[4] = { "neg", "pos", "guessed neg", "guessed pos" };
static const char *AuxLetters[4] = { "-", "+", "?-", "?+" };

static void sm(void);
static void sm_setup(void);
static int Cl(int prepos, int preneg);
static int diminishing_closure(void);
static void expand(int chosen);
static int backtrack(void);
static void backtrack_single(int apos, int reason);
static void compute_nant(void);
static int full_model(void);
static int pick_next_atom(void);
static void maybe_fire(int rule);
static void suppress(int rule, int bypos);
static void backchain_true(int rule);
static void backchain_false(int rule);
static void set_false(int atom, int chosen);
static void set_true(int atom, int chosen);
static int trypos(struct lp_atom *atom, int *SeenPos, int *SeenNeg);
static int tryneg(struct lp_atom *atom, int *SeenPos, int *SeenNeg);

static int wait_for_job(void);
static int wait_for_prog(struct lp **a_prog);

static void clear_waiting(void);
static void maybe_branch(void);
static void handle_waiting_messages(void);
static void process_stack(struct message *msg);
static void die(int v);
static int log_2(unsigned x);
static void log_stack(void);

void dump_model(void); /* for use when debugging */

static int ChoicePoints=0;
static int StableModels=0;
static int Conflicts=0;
static int ForcedAtoms=0;
static int TriedAtoms=0;

static struct lp *Prog=0;            /* The program, of course */

static struct lp_rlist **HeadOf;     /* Of which rules is this atom the head? */
static struct lp_rlist **PosDepends; /* In the bodies of which rules does this atom appear positively? */
static struct lp_rlist **NegDepends; /* In the bodies of which rules does this atom appear negatively? */

static int *PosBodySize; /* How many atoms not known true are in the positive body? */
static int *PBSize2;     /* How many atoms not known true are in the positive body? */
static int *NegBodySize; /* How many atoms not known false are in the negative body? */
static int *Suppressed;  /* How many literals are keeping this rule from firing? */
static int *HeadNumber;  /* How many nonsuppressed rules is this atom the head of? */
static int *NAnt;        /* Which atoms are negative antecedents?  Array of bools. */

static int CurrLoc;

static struct lp_aqueue *PosQueue; /* Atoms that will be put into the model */
static struct lp_aqueue *NegQueue; /* Atoms that will be removed from the model */

/* The queue for the diminishing closure is implemented as an array because it
 * is used more than the other two, and our performance depends more on it
 */
static int *DQueue;    /* Atoms that might no longer be in the upper closure */
static int DQHead;     /* Head of queue; things pulled off here */
static int DQTail;     /* Tail of queue; things tacked on here */
static int DQSize;     /* Maximum size of queue */

#define DQ_ENQUEUE(atom) do { DQueue[DQTail++]=(atom); \
                              if(DQTail>=DQSize)       \
                                      DQTail=0;        \
                              if(DQTail==DQHead) {     \
                                      DO_LOG3(LP_LOG_ERROR, "ERROR: queue overflowed (%d/%d)",DQTail,DQSize); \
                                      die(1);          \
                              }                        \
                         } while(0)

#define DQ_DEQUEUE(atom) do { atom = DQueue[DQHead++]; \
                              if(DQHead>=DQSize)       \
                                      DQHead = 0;      \
                              if(DQTail==DQHead-1) {   \
                                      DO_LOG3(LP_LOG_ERROR, "ERROR: queue underflowed (%d/%d)",DQTail,DQSize); \
                                      die(1);          \
                              }                        \
                         } while(0)

#define DQ_EMPTY (DQHead == DQTail)

#define DQ_CLEAR (DQHead = DQTail = 0)

#define DQ_ALLOC(X) (DQSize = 2*X,                   \
                     DQHead = DQTail = 0,            \
                     DQueue = lp_malloc(DQSize*sizeof(int)))



static int *Stack;     /* Stack of integers, each representing an atom */
static int *AuxStack;  /* Stack of integers, each being one of the AUX_* constants */
static int Sp;         /* Stack pointer for Stack and AuxStack:  Stack[Sp] is the first unused position */

static int *Model;     /* The current model; Model[a] is LP_L_{TRUE,FALSE,UNKNOWN} depending of atom a's value */
static int *Closure;   /* Provides an upper bound on the model.  An array of bools */
static int *Source;    /* If atom a is in the closure, Source[a] is the number of the rule which keeps a in the
			* closure */

static int UseSource=0;

#define valid_source(a,r) (!UseSource || (Source[(a)] == (r)) || Source[(a)] == -1)

static int Conflict;   /* Has a conflict been detected? */

static int ChildNumber;         /* What child is this? (shifted as we progress through InitialChoices) */
static int InitialChoices;      /* How many choices left? */


void stable_init(struct lp *prog)
{
	int i;

	begin_time(TIM_MISC, "stable_init");
	/* Allocate and initialise data structures */
	Prog = prog;

	Model   = lp_calloc(Prog->natoms,sizeof(int));
	Closure = lp_calloc(Prog->natoms,sizeof(int));
	Source = lp_malloc(Prog->natoms*sizeof(int));
	for(i=0;i<Prog->natoms;i++)
		Source[i] = -1;

	DQ_ALLOC(Prog->size);

	Stack    = lp_malloc(Prog->natoms*sizeof(int));
	AuxStack = lp_malloc(Prog->natoms*sizeof(int));
	Sp = 0;

	PosQueue = lp_aq_new();
	NegQueue = lp_aq_new();
	
	compute_nant();    /* Find and mark all negative antecedents */

	ChoicePoints = 0;
	StableModels = 0;
	Conflicts = 0;
	ForcedAtoms = 0;
	TriedAtoms = 0;
	
	Conflict = 0;
	
	/* The following functions allocate memory with lp_{c,m}alloc */
	HeadOf = lp_headof(Prog);
	HeadNumber = lp_count_head(Prog);
	PosDepends = lp_posdepends(Prog);
	NegDepends = lp_negdepends(Prog);
	PosBodySize = lp_possizes(Prog);
	PBSize2 = lp_possizes(Prog);
	NegBodySize = lp_negsizes(Prog);
	
	Suppressed = lp_calloc(Prog->nrules,sizeof(int));
	CurrLoc = 0;

	WaitingMachines = mq_new();

	/* End initialisation */
	end_time(TIM_MISC, "stable_init");
}


void stable_free(void) 
{
	int i;

	begin_time(TIM_MISC, "stable_free");
	/* Free data structures */
	lp_aq_free(PosQueue);
	lp_aq_free(NegQueue);

	for(i=0; i<Prog->natoms; i++) {
		lp_rl_free(PosDepends[i]);
		lp_rl_free(NegDepends[i]);
		lp_rl_free(HeadOf[i]);
	}
	mq_free(WaitingMachines);

	free(PosDepends);
	free(NegDepends);
	free(HeadOf);
	free(DQueue);
	free(HeadNumber);
	free(Suppressed);
	free(PosBodySize);
	free(PBSize2);
	free(NegBodySize);
	free(NAnt);
	free(Model);
	free(Closure);
	free(Source);
	free(Stack);
	free(AuxStack);
	end_time(TIM_MISC, "stable_free");	
}


int do_stable(void) 
{
	int result;
	int first=1;
	struct lp *prog;
	
	begin_time(TIM_MISC, "do_stable");
	DO_LOG(LP_LOG_DEBUG, "do_stable: calling wait_for_prog");
	if((result = wait_for_prog(&prog))) {
		DO_LOG2(LP_LOG_ERROR, "do_stable: wait_for_prog() failed (%d)", result);
		
		end_time(TIM_MISC, "do_stable_err");
		return result;
	}
	DO_LOG4(LP_LOG_INFO, "do_stable: prog=%p, prog->atoms=%p, prog->rules=%p",
		prog, prog->atoms, prog->rules);

#ifdef TRY_SINGLE_INIT
	DO_LOG(LP_LOG_DEBUG, "do_stable: initialising");
	stable_init(prog);
#endif

	do {
#ifndef TRY_SINGLE_INIT
		DO_LOG(LP_LOG_DEBUG, "do_stable: initialising");
		stable_init(prog);
#endif
		DO_LOG(LP_LOG_INFO, "do_stable: initialisation finished");
		result = 0;
		
		DO_LOG(LP_LOG_DEBUG, "do_stable: initial setup");
		/* perform the initial setup.  We have to do this before
		 * handling work, so it might as well be now.
		 */
		sm_setup();
		DO_LOG(LP_LOG_DEBUG, "do_stable: initial setup finished");
		
		if(!first) {
			DO_LOG(LP_LOG_DEBUG, "do_stable: calling wait_for_job()");
			result = wait_for_job();
			DO_LOG(LP_LOG_DEBUG, "do_stable: wait_for_job() finished");
		} else {
			first = 0;
			sm();
		}
		if(result == 0)	{
			DO_LOG(LP_LOG_DEBUG, "do_stable: calling send_waiting()");
			send_waiting(pvm_parent());
			DO_LOG(LP_LOG_DEBUG, "do_stable: send_waiting() returned");
		}

#ifndef TRY_SINGLE_INIT
		DO_LOG(LP_LOG_DEBUG, "do_stable: freeing structures");
		stable_free();
		DO_LOG(LP_LOG_DEBUG, "do_stable: freed");
#else
		memset(Model, 0, Prog->natoms*sizeof(int));
		memset(Closure, 0, Prog->natoms*sizeof(int));
		memset(Source, -1, Prog->natoms*sizeof(int));
		memcpy(PBSize2, PosBodySize, Prog->nrules*sizeof(int));
		DQ_CLEAR;
		Sp = 0;
#endif
		
	} while(result==0);
	
#ifdef TRY_SINGLE_INIT
	DO_LOG(LP_LOG_DEBUG, "do_stable: freeing structures");
	stable_free();
	DO_LOG(LP_LOG_DEBUG, "do_stable: freed");
#endif
	end_time(TIM_MISC, "do_stable");
	send_died(pvm_parent());
	return result;
}


int wait_for_prog(struct lp **a_prog)
{
	struct message m;
	int result;
	int mac;
	int lg;

	begin_time(TIM_MISC, "wait_for_prog");
	
	/* Make sure we are notified of our parent's death */
	mac = pvm_parent();
	DO_LOG2(LP_LOG_DEBUG, "wait_for_prog: calling pvm_notify() for parent (%x)", mac);
	if((result = pvm_notify(PvmTaskExit, MTAG_NOTIFY, 1, &mac)))
		DO_LOG2(LP_LOG_ERROR, "wait_for_prog: pvm_notify() failed (%d)", result);
	else
		DO_LOG(LP_LOG_DEBUG, "wait_for_prog: pvm_notify() succeeded");
	
	while(1) {
		DO_LOG(LP_LOG_DEBUG, "wait_for_prog: calling wait_for_message()");
		if((result = wait_for_message(&m))) {
			DO_LOG2(LP_LOG_ERROR, "wait_for_prog: wait_for_message() failed (%d)", result);
			goto done;
		}
		DO_LOG(LP_LOG_DEBUG, "wait_for_prog: wait_for_message() succeeded");
		
		switch(m.type) {
		    case MSG_PROG:
			DO_LOG(LP_LOG_DEBUG, "wait_for_prog: got MSG_PROG");
			*a_prog = (struct lp *)m.data;
			
			ChildNumber = m.misc;
			lg = log_2(m.misc2-1);

			/* Note that log_2 rounds up.  We wish to make log_2 choices only if
			 * we have a sibling which matches our last choice.  Consider, for
			 * example, the case of 11 children (log_2(11) = 4):
			 *   child[0] = 0000         child[8]  = 1000
			 *   child[1] = 0001         child[9]  = 1001
			 *   child[2] = 0010         child[10] = 1010
			 *   child[3] = 0011
			 *   child[4] = 0100
			 *   child[5] = 0101
			 *   child[6] = 0110
			 *   child[7] = 0111
			 *
			 * child[0] should assume false, false, false, false, because child[8]
			 * will be assuming false, false, false, true (proceed from right to
			 * left).  However, child[6] should assume only false, true, true,
			 * since if it guesses an additional false, there will be no other
			 * child to guess true.  So children 3 (= 11 - 8) through 7 (= 8 - 1)
			 * should have their number of assumptions reduced by one.
			 */
			if(m.misc >= (m.misc2 - (1 << (lg-1))) && m.misc < (1 << (lg-1)))
				InitialChoices = lg-1;
			else
				InitialChoices = lg;
			
			DO_LOG4(LP_LOG_INFO, "wait_for_prog: I am child %d/%d (%d choices)", m.misc, m.misc2, InitialChoices);
			
			result = 0;
			goto done;
			
		    case MSG_DIE:
			DO_LOG2(LP_LOG_INFO, "wait_for_prog: got MSG_DIE from %x", m.from);
			DO_LOG2(LP_LOG_ERROR, "dying at request of %x", m.from);
			die(0);
		    case MSG_DIED:
			if(m.from == pvm_parent()) {
				DO_LOG(LP_LOG_ERROR, "wait_for_prog: parent died---I should too");
				die(1);
			}
			break;
		    case MSG_STACK:
		    case MSG_STACK_UPD:
			free(m.data);
			/* FALL-THROUGH */
		    default:
			not_expecting(m.type);
		}
	}
 done:
	end_time(TIM_MISC, result ? "wait_for_prog_err" : "wait_for_prog");
	return result;
}


/* stable_init has already been called */
int wait_for_job(void)
{
	struct message m;
	int result=0;
	begin_time(TIM_MISC, "wait_for_job");
	
	do {
		DO_LOG(LP_LOG_DEBUG, "wait_for_job: calling wait_for_message()");
		if((result = wait_for_message(&m))) {
			DO_LOG2(LP_LOG_ERROR, "wait_for_job: wait_for_message() failed (%d)", result);
			goto done;
		}
		DO_LOG(LP_LOG_DEBUG, "wait_for_job: wait_for_message() succeeded");
		
		switch(m.type) {
		    case MSG_STACK:
		    case MSG_STACK_UPD:
			DO_LOG3(LP_LOG_INFO, "wait_for_job: got MSG_STACK%s from %x",
				m.type==MSG_STACK_UPD ? "_UPD" : "", m.from);
			process_stack(&m);
			
			break;
		    case MSG_NAKWORK:
			DO_LOG2(LP_LOG_INFO, "wait_for_job: got MSG_NAKWORK from %x", m.from);
			result=0;
			goto done; /* get work somewhere else */
		    case MSG_WAITING:
			DO_LOG2(LP_LOG_INFO, "wait_for_job: got MSG_WAITING from %x", m.from);
			/* I have no work, so why bother? */
			send_nakwork(m.from);
			break;
		    case MSG_DIE:
			DO_LOG2(LP_LOG_INFO, "wait_for_job: got MSG_DIE from %x", m.from);
			DO_LOG2(LP_LOG_ERROR, "dying at request of %x", m.from);
			die(0);
		    case MSG_DIED:
			if(m.from == pvm_parent()) {
				DO_LOG(LP_LOG_ERROR, "wait_for_job: parent died---I should too");
				die(1);
			}
			break;
		    default:
			not_expecting(m.type);
		}
	} while(m.type != MSG_STACK && m.type != MSG_STACK_UPD);

	DO_LOG(LP_LOG_DEBUG, "wait_for_job: expand()ing from branched stack");
	if(!Verbose_timing) /* if Verbose_timing is on, expand() will call this */
		begin_time(TIM_COMPUTE, "expand");
	expand(0);
	if(!Verbose_timing) /* if Verbose_timing is on, expand() will call this */
		end_time(TIM_COMPUTE, "expand");
	DO_LOG3(LP_LOG_DEBUG, "wait_for_job: expand() returned, Sp=%d, Conflict=%d", Sp, Conflict);

	DO_LOG(LP_LOG_DEBUG, "wait_for_job: calling sm()");
	sm();
	result = 0;
 done:
	end_time(TIM_MISC, "wait_for_job");
	return result;
}


void process_stack(struct message *msg)
{
	int i;
	int val, apos;
	int *dat = (int *)msg->data;
	int sp = msg->size;
	char pbuf[80];
	
	sprintf(pbuf, "process_stack from %x, %d atoms", msg->from, msg->size);	
	begin_time(TIM_COMPUTE, "process_stack");

	/* XXX: try putting them on the queue instead? */

	for(i=0; i<sp; ++i) {
		apos = dat[i];
		val = dat[i + sp];

		if(Model[apos]) {
			if((Model[apos] == LP_L_TRUE  && val == AUX_EXP_NEG) ||
			   (Model[apos] == LP_L_FALSE && val == AUX_EXP_POS)) {
				DO_LOG3(LP_LOG_ERROR, "process_stack: ERROR: Conflict on atom %d (%s)",
					apos, Prog->atoms[apos].name);
				DO_LOG3(LP_LOG_ERROR, "process_stack: model = %s, stack = %s",
					Model[apos] == LP_L_TRUE ? "True" : "False", AuxNames[val]);
				Conflict = 1;
			} else {
				/* val corresponds correctly */
				DO_LOG4(LP_LOG_DEBUG, "process_stack: atom %d (%s) already in model (%s)---skipping",
					apos, Prog->atoms[apos].name,
					Model[apos] == LP_L_TRUE ? "True" : "False");
			}
		} else {
			DO_LOG4(LP_LOG_DEBUG, "process_stack: setting %d (%s) to %s", apos,
				Prog->atoms[apos].name, val ? "true" : "false");

			if(val == AUX_EXP_POS)
				set_true(apos,0);
			else if(val == AUX_EXP_NEG)
				set_false(apos,0);
			else {
				DO_LOG2(LP_LOG_ERROR, "process_stack: ERROR: bad value %d", val);
				die(1);
			}
		}
	}
	
	DO_LOG3(LP_LOG_DEBUG, "process_stack: sp = %d, Sp = %d", sp, Sp);
	free(dat);
	end_time(TIM_COMPUTE, "process_stack");
}


/* Perform the computation.  Assumes all appropriate variables and data
 * structures have been initialised by lp_stable().
 */
void sm(void) 
{
	static int i=0;
	
	begin_time(TIM_COMPUTE, "sm");
	while(1) {
		if(Conflict) {
			++Conflicts;
			DO_LOG3(LP_LOG_DEBUG, "sm: got a conflict (number %d) [Sp=%d]", Conflicts, Sp);
			Conflict = 0;
			
			if(backtrack()) {
				DO_LOG(LP_LOG_DEBUG, "sm: backtracked all the way");
				goto bt_done;
			}
			/* note: InitialChoices must be zero if we get here */
			
			DO_LOG2(LP_LOG_DEBUG, "sm: continuing from backtrack [Sp=%d]", Sp);
			continue;
		} else if(full_model()) {
			++StableModels;
			DO_LOG3(LP_LOG_INFO, "sm: got a model (number %d) [Sp=%d]", StableModels, Sp);

			/* ChildNumber should be 0 when we have exhausted InitialChoices */
			if(ChildNumber != 0) {
				DO_LOG(LP_LOG_INFO, "sm: in common area of tree; will be handled by other child");
			} else {
				int result;
				
				if(InitialChoices > 0) {
					DO_LOG(LP_LOG_INFO, "sm: in common area of tree; I am lowest child, reporting");
				}

				DO_LOG(LP_LOG_DEBUG, "have model!!");
				log_stack();

				DO_LOG(LP_LOG_DEBUG, "sm: calling send_model()");
				if((result = send_model_CH(pvm_parent(), Prog, Model))) {
					DO_LOG2(LP_LOG_ERROR, "sm: send_model() failed (%d)", result);
					goto bt_done;
				}
				DO_LOG(LP_LOG_DEBUG, "sm: send_model() succeeded");
			}
			
			/* Nothing is unknown, so backtrack */
			if(backtrack()) {
				DO_LOG(LP_LOG_DEBUG, "sm: backtracked all the way");
				goto bt_done;
			}
			/* note: InitialChoices must be zero if we get here */
			
			DO_LOG2(LP_LOG_DEBUG, "sm: continuing from backtrack [Sp=%d]", Sp);
			continue;
		}

		/* We cannot expand anything else, so lookahead */
		if(pick_next_atom()) {
			/* we had to guess */
			++ChoicePoints;
			DO_LOG2(LP_LOG_DEBUG, "sm: choice point number %d", ChoicePoints);
			DO_LOG(LP_LOG_DEBUG, "sm: calling handle_waiting_messages()");
			handle_waiting_messages();
			DO_LOG(LP_LOG_DEBUG, "sm: handle_waiting_messages() returned");
			DO_LOG(LP_LOG_DEBUG, "sm: calling maybe_branch()");
			maybe_branch();
			DO_LOG(LP_LOG_DEBUG, "sm: maybe_branch() returned");
		} else {
			/* something caused a conflict, so we added its negation */
			++ForcedAtoms;
			DO_LOG2(LP_LOG_DEBUG, "sm: forced atom number %d", ForcedAtoms);
			
			/* XXX: should we do this?  if so, perhaps more/less often? */
			if(CheckFreq && ((++i)%CheckFreq == 0)) {
				handle_waiting_messages();
				maybe_branch();
			}
		}
	}

	DO_LOG(LP_LOG_ERROR, "sm: can't happen---shouldn't leave sm() loop");
	end_time(TIM_COMPUTE, "sm_cant_happen");
	return; /* Can't happen */
	
	/* clear out WaitingMachines */
 bt_done:
	/* Once we backtrack fully, we are finished with our initial subtree,
	 * so we should no longer make choices as though we were still there.
	 */
	InitialChoices = 0;

	clear_waiting();
	
	end_time(TIM_COMPUTE, "sm_work_done");
	return;
}

void clear_waiting(void) 
{
	DO_LOG(LP_LOG_INFO, "sm: no more work---sending MSG_NAKWORK to waiting machines");
	while(!mq_empty(WaitingMachines)) {
		machine_t machine;
		machine = mq_dequeue(WaitingMachines);
		if(machine == pvm_mytid()) {
			
			DO_LOG(LP_LOG_ERROR, "sm: ERROR: I was in my own WaitingMachines queue");
			continue;
		}
		DO_LOG2(LP_LOG_INFO, "sm: calling send_nakwork() to %x", machine);
		send_nakwork(machine);
	}
}


/* Perform the initial expansion.  Basically, this function sets Model to the WFS of Prog, and
 * sets Closure to Dcl(P_{WFS(P)},not(NAnt(P))).
 */
void sm_setup(void) 
{
	int i, apos;

	begin_time(TIM_COMPUTE, "sm_setup");
	
	/* First, handle negative requirements */
	DO_LOG2(LP_LOG_DEBUG, "sm_setup: Handling %d negative requirements", Prog->nnegreq);
	for(i=0; i<Prog->nnegreq; i++) {
		set_false(lp_apos(Prog,Prog->negreq[i]),0);
	}
	/* Then positive ones */
	DO_LOG2(LP_LOG_DEBUG, "sm_setup: Handling %d positive requirements", Prog->nnegreq);
	for(i=0; i<Prog->nposreq; i++) {
		set_true(lp_apos(Prog,Prog->posreq[i]),0);
	}
	
	/* Then, enqueue the heads of those rules which must fire */
	for(i=0; i<Prog->nrules; i++) {
		if(Prog->rules[i].npos == 0) {
			apos = lp_apos(Prog,Prog->rules[i].head);

			DO_LOG4(LP_LOG_DEBUG, "sm_setup: putting head atom %d (%s) of rule %d into closure",
				apos, Prog->rules[i].head->name, i);
			
			DQ_ENQUEUE(apos);
			Source[apos] = i;
			
			if(Prog->rules[i].nneg == 0) {
				DO_LOG(LP_LOG_DEBUG, "sm_setup: must fire---also putting into model");
				lp_aq_enqueue(PosQueue,Prog->rules[i].head);
			}
		}
	}

	/* Now, find things that should be in the diminishing closure */
	while(!DQ_EMPTY) {
		DQ_DEQUEUE(apos);

		/* If we haven't processed this atom yet */
		if(Closure[apos] == 0) {
			struct lp_rlist *j;

			for(j=PosDepends[apos]; j!=NULL; j=j->next) {
				int rpos = lp_rpos(Prog,j->rule);
				
				if(--PBSize2[rpos] == 0 && !Suppressed[rpos]) {
					int hpos = lp_apos(Prog,j->rule->head);
					DO_LOG4(LP_LOG_DEBUG, "sm_setup: putting head atom %d (%s) of rule %d "
						"into closure", hpos, j->rule->head->name, rpos);
					DQ_ENQUEUE(hpos);
					Source[hpos] = rpos;
				}
			}
			Closure[apos] = 1;
		}
	}

	/* If something is not in the diminishing closure, it cannot be in
	 * the model, either.
	 */
	for(i=0; i<Prog->natoms; i++) {
		if(Closure[i])
			continue;
		if(Model[i] == LP_L_TRUE) {
			Conflict = 1;
			Conflicts++;
			break;
		}
		DO_LOG3(LP_LOG_DEBUG, "sm_setup: atom %d (%s) cannot be true---setting false",
			i, Prog->atoms[i].name);
		set_false(i,0);
	}

	/* Deal with everything that has been added to the positive queue. */
	if(!Conflict) {
		DO_LOG(LP_LOG_DEBUG, "sm_setup: calling expand()");
		expand(0);
		DO_LOG2(LP_LOG_INFO, "sm_setup: expand() returned, Sp=%d", Sp);
	}

	end_time(TIM_COMPUTE, "sm_setup");
}


/* Backtrack from a single expansion/choice.  This function reverts all changes
 * made when apos was pushed onto the stack.  reason should be the value of
 * AuxStack that goes along with apos.  This function should only be called on
 * the top atom of the stack (or an atom that was just popped off the stack).
 */
void backtrack_single(int apos, int reason) 
{
	struct lp_rlist *j;
	int rpos, hpos;
	
	if(Verbose_timing)
		begin_time(TIM_COMPUTE, "backtrack_single");
	DO_LOG4(LP_LOG_EXTRA, "backtrack_single: backtracking atom %d (%s) [%s]",
		apos, Prog->atoms[apos].name, AuxNames[reason]);

	switch(reason) {
	    case AUX_POS:
	    case AUX_EXP_POS:
		/* Inactivate rules apos caused to fire */
		for(j=PosDepends[apos]; j!=NULL; j=j->next) {
			rpos = lp_rpos(Prog,j->rule);
			++PosBodySize[rpos];
		}
		/* Re-enable rules apos suppressed */
		for(j=NegDepends[apos]; j!=NULL; j=j->next) {
			rpos = lp_rpos(Prog,j->rule);
			hpos = lp_apos(Prog,j->rule->head);

			if(--Suppressed[rpos] == 0)
				HeadNumber[hpos]++;
		}
		break;
			
	    case AUX_NEG:
	    case AUX_EXP_NEG:
		/* Inactivate rules we caused to fire */
		for(j=NegDepends[apos]; j!=NULL; j=j->next) {
			rpos = lp_rpos(Prog,j->rule);
			++NegBodySize[rpos];
		}
		/* Re-enable rules apos suppressed */
		for(j=PosDepends[apos]; j!=NULL; j=j->next) {
			rpos = lp_rpos(Prog,j->rule);
			hpos = lp_apos(Prog,j->rule->head);

			--PBSize2[rpos];
			
			if(--Suppressed[rpos] == 0)
				HeadNumber[hpos]++;
		}
		break;
			
	    default:
		DO_LOG4(LP_LOG_ERROR, "backtrack_single: ERROR: bad value on AuxStack: %d, atom %d (%s)",
			reason, apos, Prog->atoms[apos].name);
		die(1);
	}

	/* Put this atom back in the diminishing closure; the next expand()
	 * will take it out if need be.
	 */
	Closure[apos] = 1;
	Source[apos] = -1;
	
	Model[apos] = LP_L_UNKNOWN;
	if(Verbose_timing)
		end_time(TIM_COMPUTE, "backtrack_single");
}


/* Pop literals off the stack until we reach a chosen (rather than forced
 * or expanded) atom; then pop off that literal and expand on its negation.
 *
 * This function returns 1 if the search space has been exhausted, 0
 * otherwise.
 */
int backtrack(void) 
{
	int apos;
	int reason;
	int replace;
	int result=0;

	if(Verbose_timing)
		begin_time(TIM_COMPUTE, "backtrack");
	DO_LOG2(LP_LOG_DEBUG, "backtrack: backtracking from Sp=%d", Sp);
	
	while(Sp > 0) {
		--Sp;
		apos = Stack[Sp];
		reason = AuxStack[Sp];
		replace = 0;
		Closure[apos] = 0;

		backtrack_single(apos, reason);

		/* If the atom was chosen, enqueue its negation and expand */
		switch(reason) {
		    case AUX_EXP_POS:
		    case AUX_EXP_NEG:
			break;
		    case AUX_POS:
			lp_aq_enqueue(NegQueue,&Prog->atoms[apos]);
			DO_LOG2(LP_LOG_DEBUG, "backtrack: finished backtracking at Sp=%d", Sp);
			DO_LOG3(LP_LOG_DEBUG, "backtrack: making atom %d (%s) negative and expanding", apos,
				Prog->atoms[apos].name);
			expand(0);
			DO_LOG2(LP_LOG_DEBUG, "backtrack: finished expanding [Sp=%d]", Sp);
			
			goto done;
		    case AUX_NEG:
			lp_aq_enqueue(PosQueue,&Prog->atoms[apos]);
			DO_LOG2(LP_LOG_DEBUG, "backtrack: finished backtracking at Sp=%d", Sp);
			DO_LOG3(LP_LOG_DEBUG, "backtrack: making atom %d (%s) positive and expanding", apos,
				Prog->atoms[apos].name);
			expand(0);
			DO_LOG2(LP_LOG_DEBUG, "backtrack: finished expanding [Sp=%d]", Sp);

			goto done;
		    default:
			/* backtrack_single would have logged an error */
			DO_LOG(LP_LOG_ERROR, "backtrack: can't happen: bad value on AuxStack");
			die(1);
		}
	}	

	DO_LOG(LP_LOG_DEBUG, "backtrack: hit bottom");
	result=1;
 done:
	if(Verbose_timing)
		end_time(TIM_COMPUTE, "backtrack");
	return result;
}


/* Assuming Conflict is false, returns 1 if Model is a stable model, otherwise
 * 0.  Should not be called if Conflict is true.
 */
int full_model(void) 
{
	return Sp >= Prog->natoms;
}


/* Compute the set of negative antecedents of Prog.  Fills NAnt[]. */
void compute_nant(void) 
{
	int i,j, apos;
	struct lp_rule *r;
	begin_time(TIM_COMPUTE, "compute_nant");
	NAnt = lp_calloc(Prog->natoms,sizeof(int));
	
	for(i=0; i<Prog->nrules; i++) {
		r = &Prog->rules[i];
		for(j=0;j<r->nneg;j++) {
			apos = lp_apos(Prog,r->neg[j]);
			NAnt[apos] = 1;
		}
	}
	end_time(TIM_COMPUTE, "compute_nant");
}


/* Expand on the set of literals in PosQueue, NegQueue, and DQueue.
 * DQueue is usually empty when expand() is called.  If chosen is true,
 * backtrack() will try the negations of these literals . Otherwise, they
 * are assumed to be forced, and backtrack() will simply backtrack past
 * them.
 */
void expand(int chosen) 
{
	int possize,negsize, junk;
	if(Verbose_timing)
		begin_time(TIM_COMPUTE, "expand");
	
	if(chosen) {
		possize=lp_aq_size(PosQueue);
		negsize=lp_aq_size(NegQueue);
	} else {
		possize = 0;
		negsize = 0;
	}
	

	if(Cl(possize,negsize)) {
		/* We have a conflict: empty the queues */
		while(!lp_aq_empty(PosQueue))
			(void)lp_aq_dequeue(PosQueue);
		while(!lp_aq_empty(NegQueue))
			(void)lp_aq_dequeue(NegQueue);
		while(!DQ_EMPTY)
			DQ_DEQUEUE(junk);
		
	}

	if(Verbose_timing)
		end_time(TIM_COMPUTE, "expand");
}


/* Perform the actual expansion.  prepos is the number of chosen (as opposed
 * to expanded or forced) positive literals to push; preneg is the number of
 * chosen negative literals to push.  Returns 1 if there was a conflict,
 * zero otherwise.
 */
int Cl(int prepos, int preneg)
{
	struct lp_atom *a;
	int apos;
	int result=0;

	if(Verbose_timing)
		begin_time(TIM_COMPUTE, "Cl");
	for(;;) {
		if(Conflict) {
			result=1;
			goto done;
		}
		
		if(!lp_aq_empty(PosQueue)) {
			/* Make something positive */
			a = lp_aq_dequeue(PosQueue);
			apos = lp_apos(Prog,a);
			if(Model[apos] == LP_L_FALSE)
				Conflict = 1;
			else if(!Model[apos])
				set_true(apos, (prepos-- > 0));
		} else if(!lp_aq_empty(NegQueue)) {
			/* Make something negative */
			a = lp_aq_dequeue(NegQueue);
			apos = lp_apos(Prog,a);
			if(Model[apos] == LP_L_TRUE)
				Conflict = 1;
			else if(!Model[apos])
				set_false(apos, (preneg-- > 0));
		} else if(DQ_EMPTY) {
			/* The queues are empty, we are finished */
			goto done;
		} else {
			/* Possibly remove things from the diminishing closure */
			if(diminishing_closure())
				Conflict = 1;
		}
	}
 done:
	if(Verbose_timing)
		end_time(TIM_COMPUTE, "Cl");
	return result;
}


/* For each atom in DQueue, remove it from the diminishing closure
 * if necessary.  Returns true if there was a conflict, false otherwise.
 *
 * TODO:  Add source pointers a la smodels 2, to improve performance.
 */
int diminishing_closure(void) 
{
	/* TODO: this is the single greatest source of calls to
	 * lp_apos() and lp_rpos().  Source pointers may help, but
	 * what else?
	 */
	/* struct lp_atom *a; */
	struct lp_rlist *j;
	int apos, rpos, hpos, i;
	
	/* Queue of atoms temporarily assumed to be out of the closure */
	int *tmpf = lp_calloc(Prog->natoms, sizeof(int));
	/* Tail of tmpf: tmpf[tp] is one past the last atom enqueued */
	int tp = 0;

	int result=0;

	if(Verbose_timing)
		begin_time(TIM_COMPUTE, "diminishing_closure");
	
	/* First, assume everything in DQueue is out of the closure */
	while(!DQ_EMPTY) {
		DQ_DEQUEUE(apos);

		/* If it was previously in the closure, and it was not in the
		 * model negatively, take it out of the closure and add it to
		 * tmpf.
		 */
		if(Closure[apos] == 1 && Model[apos] != LP_L_FALSE) {
			for(j=PosDepends[apos]; j!=NULL; j=j->next) {
				rpos = lp_rpos(Prog, j->rule);
				hpos = lp_apos(Prog, j->rule->head);
				if(++PBSize2[rpos] == 1 && !Suppressed[rpos] && Closure[hpos] && valid_source(hpos,rpos))
					DQ_ENQUEUE(lp_apos(Prog,j->rule->head));
			}
			Closure[apos] = 0;
			tmpf[tp++] = apos; /* enqueue */
		}
	}
	
	/* Enqueue everything which was taken out but should still be in. */
	for(i=0; i<tp; i++) {
		for(j=HeadOf[tmpf[i]]; j!=NULL; j=j->next) {
			rpos = lp_rpos(Prog, j->rule);
			if(PBSize2[rpos] == 0 && !Suppressed[rpos]) {
				hpos = lp_apos(Prog,j->rule->head);
				
				Source[hpos] = rpos;
				DQ_ENQUEUE(hpos);
				break;
			}
		}
	}

	/* Propagate truth: add everything in the queue, and everything that
	 * follows from those things, back to the closure.
	 */
	while(!DQ_EMPTY) {
		DQ_DEQUEUE(apos);
		
		if(Closure[apos] == 0 && Model[apos] != LP_L_FALSE) {
			for(j=PosDepends[apos]; j!=NULL; j=j->next) {
				/* propagateTrue */
				rpos = lp_rpos(Prog, j->rule);
				hpos = lp_apos(Prog, j->rule->head);
				if(--PBSize2[rpos] == 0 && !Suppressed[rpos] && !Closure[hpos]) {
					hpos = lp_apos(Prog,j->rule->head);

					Source[hpos] = rpos;
					DQ_ENQUEUE(hpos);
				}
			}
			Closure[apos] = 1;
		}
	}

	/* If something is not in the closure, it must be false */
	for(i=0; i<tp; i++) {
		apos = tmpf[i];
		if(!Closure[apos]) {
			if(Model[apos] == LP_L_TRUE) {
				/* Revert changes made by diminishing_closure */
				for(; i < tp; i++) {
					apos = tmpf[i];
					if(!Closure[apos]) {
						for(j=PosDepends[apos]; j!=NULL; j=j->next)
							--PBSize2[lp_rpos(Prog,j->rule)];
						/* Source[apos] should be as it was */
						Closure[apos] = 1;
					}
				}
				
				Conflict = 1;
				result=1;
				goto done;
			} else {
				set_false(apos, 0);
				
				/* Revert changes made by diminishing_closure */
				for(j=PosDepends[apos]; j!=NULL; j=j->next)
					--PBSize2[lp_rpos(Prog,j->rule)];
			}
		}
	}

 done:
	free(tmpf);

	if(Verbose_timing)
		end_time(TIM_COMPUTE, "diminishing_closure");
	return result;
}


/* Possibly fire a rule.  Called whenever a literal in the rules body is added to the
 * model.  If all literals in the body are satisfied, enqueues the head to be added
 * to the model.  If the head is known false and only one body literal is not satisfied,
 * the opposite of that literal is enqueued.
 */
void maybe_fire(int rule) 
{
	int hpos;
	if(Suppressed[rule])
		return;
	hpos = lp_apos(Prog,Prog->rules[rule].head);
	if(PosBodySize[rule] == 0 && NegBodySize[rule] == 0 && Model[hpos] != LP_L_TRUE)
		lp_aq_enqueue(PosQueue, Prog->rules[rule].head);
	else if(NegBodySize[rule] + PosBodySize[rule] == 1 && Model[hpos] == LP_L_FALSE)
		backchain_false(rule);
}


/* Suppress a rule.  bypos is true if the rule is being suppressed because an atom
 * in the rule's positive body was added negatively to the model.  Decrements 
 * HeadNumber[] for the head of the rule, possibly adding the head's negation to
 * the model or backchaining from the head.
 */
void suppress(int rule, int bypos) 
{
	int apos,rpos;
	struct lp_rlist *j;

	if(bypos)
		++PBSize2[rule];
	
	if(++Suppressed[rule] == 1) {
		apos = lp_apos(Prog,Prog->rules[rule].head);
		if (--HeadNumber[apos] != 0 && Model[apos] != LP_L_FALSE)
			DQ_ENQUEUE(apos);
		
		if(Model[apos] != LP_L_FALSE && HeadNumber[apos] == 0)
			lp_aq_enqueue(NegQueue,&Prog->atoms[apos]);
		else if(Model[apos] == LP_L_TRUE && HeadNumber[apos] == 1) {
			for(j=HeadOf[apos]; j!=NULL; j=j->next) {
				rpos = lp_rpos(Prog,j->rule);
				if(Suppressed[rpos] == 0) {
					backchain_true(rpos);
					break;
				}
			}
		}
	}
}


/* If we know a rule's head is true, and that atom is the head of this rule
 * only (as well as, possibly, suppressed rules), then all the literals in
 * the rule's body must be satisfied.  This function does that.  It should
 * only be called if it is known that the head of this rule is the head of
 * this rule and suppressed rules only, and that this rule is unsuppressed.
 */
void backchain_true(int rule)  
{
	int i;
	struct lp_rule *r = &Prog->rules[rule];
	
	if(Model[lp_apos(Prog,r->head)] == LP_L_TRUE) {
		/* Add each body literal to the appropriate queue */
		for(i=0; i<r->npos; i++)
			lp_aq_enqueue(PosQueue,r->pos[i]);
		for(i=0; i<r->nneg; i++)
			lp_aq_enqueue(NegQueue,r->neg[i]);
	}
}


/* If we know a rule's head is false, and that there is only one unsatisfied
 * literal in the rule's body, then the negation of that literal must be in
 * the model.  This function adds that literal's negation.  It should only be
 * called if the head of this rule is known to be false.
 */
void backchain_false(int rule) 
{
	int apos;
	int i;
	struct lp_rule *r = &Prog->rules[rule];
	
	if(PosBodySize[rule] == 0 && NegBodySize[rule] == 1) {
		/* find the unsatisfied negative literal */
		for(i=0; i<r->nneg; i++) {
			apos = lp_apos(Prog,r->neg[i]);
			if(!Model[apos]) {
				lp_aq_enqueue(PosQueue,r->neg[i]);
				break;
			}
		}
	} else if(PosBodySize[rule] == 1 && NegBodySize[rule] == 0) {
		/* find the unsatisfied positive literal */
		for(i=0; i<r->npos; i++) {
			apos = lp_apos(Prog,r->pos[i]);
			if(!Model[apos]) {
				lp_aq_enqueue(NegQueue,r->pos[i]);
				break;
			}
		}
	}
}


/* Assuming we have expanded fully, choose the next literal to try adding to
 * the model.  Only literals in the program's negative antecedent are tried.
 * If some literal causes a conflict, its negation is added and this function
 * returns false; this is known as `forcing'.  Otherwise, we choose an atom
 * which maximises the number of atoms added in the worst case, add that atom
 * (possibly negated) to the model, and return true; this is known as
 * `choosing': decisions made by choosing may be reversed in backtracking.
 *
 * As an optimisation, we note that, if adding the literal A and expanding also
 * adds B, C, and D, then we know that adding B, C, or D cannot result in a
 * larger expansion than that produced by expanding A.
 */
int pick_next_atom(void) 
{
	int apos;

	/* True if expanding some other literal resulted in this one */
	int *SeenPos = lp_calloc(Prog->natoms, sizeof(int));
	int *SeenNeg = lp_calloc(Prog->natoms, sizeof(int));
	
	int foundone=0;
	int hi_min=-1, hi_max=-1;
	int hi_index = -1;
	int hi_pos = 0;
	int pos, neg, min, max;
	int result=0;

	if(Verbose_timing)
		begin_time(TIM_COMPUTE, "pick_next_atom");
	
	/* if Prog->natoms = 0, full_model() should have caught it in sm() */
	if(Prog->natoms == 0) {
		DO_LOG(LP_LOG_ERROR, "pick_next_atom(): ERROR: CALLED WITH Prog->natoms=0\n");
		die(1);
	}

	if(CurrLoc == Prog->natoms)
		CurrLoc = 0;
	
	apos = CurrLoc;
	
	while(1) {
		pos = 0;
		neg = 0;
		
		/* FIXME: we might be able to know for sure that this won't work
		 * even if only one has been seen */

		/* if it is a negative antecedent, unknown, and
		 *   * may cause conflict
		 *     OR
		 *   * may have a better heuristic.
		 */
		if(Model[apos]==LP_L_UNKNOWN
		   && (!SeenPos[apos] || !SeenNeg[apos])) {
			foundone = 1;

			if(!SeenPos[apos]) {
				if((pos = trypos(&(Prog->atoms[apos]), SeenPos, SeenNeg)) < 0) {
					DO_LOG4(LP_LOG_DEBUG, "pick_next_atom: forcing atom %d (%s) to be "
						"negative [Sp=%d]", apos, Prog->atoms[apos].name,
						Sp);
					CurrLoc = apos+1;
					goto done;
				}
			}
			if(!SeenNeg[apos]) {
				if((neg = tryneg(&(Prog->atoms[apos]), SeenPos, SeenNeg)) < 0) {
					DO_LOG4(LP_LOG_DEBUG, "pick_next_atom: forcing atom %d (%s) to be "
						"positive [Sp=%d]", apos, Prog->atoms[apos].name,
						Sp);
					CurrLoc = apos+1;
					goto done;
				}
			}
			if(pos > neg) {
				min = neg;
				max = pos;
			} else {
				min = pos;
				max = neg;
			}
			
			/* If this atom has the best minimum and the best maximum so far */
			if(min > hi_min || (min == hi_min && max > hi_max)) {
				hi_min = min;
				hi_max = max;
				hi_index = apos;
				hi_pos = (max == pos);
			}
		}
		
		if(++apos == Prog->natoms)
			apos=0;
				
		if(apos == CurrLoc)
			break;
	}

	/* If we are still in the shared part of the tree, make the appropriate choice */
	if(InitialChoices) {
		--InitialChoices;
		lp_aq_enqueue(ChildNumber % 2 ? PosQueue : NegQueue, &Prog->atoms[hi_index]);
		DO_LOG4(LP_LOG_DEBUG, "pick_next_atom: setting atom %d (%s) = %s (initial)", hi_index,
			Prog->atoms[hi_index].name, ChildNumber & 0x1 ? "true" : "false");
		TriedAtoms++;
		expand(0); /* not being guessed---some other machine will handle the opposite case */
		
		ChildNumber >>= 1;

		goto done;
	}
	
	lp_aq_enqueue(hi_pos ? PosQueue : NegQueue, &Prog->atoms[hi_index]);
	DO_LOG4(LP_LOG_DEBUG, "pick_next_atom: guessing atom %d (%s) to be %s", hi_index,
		Prog->atoms[hi_index].name, hi_pos ? "positive" : "negative");
	TriedAtoms++;
	expand(1);
	DO_LOG2(LP_LOG_DEBUG, "pick_next_atom: Sp=%d", Sp);
	result = 1;
	
 done:
	free(SeenPos);
	free(SeenNeg);
	if(Verbose_timing)
		end_time(TIM_COMPUTE, "pick_next_atom");
	return result;
}


/* Set an atom to false, and propagate that information.  If chosen is true,
 * backtracking will try this atom as true.
 */
void set_false(int atom, int chosen) 
{
	struct lp_rlist *j;
	int rpos;

	Model[atom] = LP_L_FALSE;
	Closure[atom] = 0;

	/* push this atom */
	Stack[Sp] = atom;
	AuxStack[Sp++] = chosen ? AUX_NEG : AUX_EXP_NEG;

	/* No rule with this atom in its positive body can fire */
	for(j=PosDepends[atom]; j!=NULL; j=j->next) {
		rpos = lp_rpos(Prog,j->rule);
		suppress(rpos,1);
	}
	/* We have put some rules closer to firing */
	for(j=NegDepends[atom]; j!=NULL; j=j->next) {
		rpos = lp_rpos(Prog,j->rule);
		--NegBodySize[rpos];
		maybe_fire(rpos);
	}
	/* If this atom is still the head of some rules, see if we can find
	 * out anything about the truth values for the bodies of those rules.
	 */
	if(HeadNumber[atom] != 0) {
		for(j=HeadOf[atom]; j!=NULL; j=j->next) {
			rpos = lp_rpos(Prog,j->rule);
			backchain_false(rpos);
		}
	}
}


/* Set an atom to true, and propagate that information.  If chosen is true,
 * backtracking will try this atom as false.
 */
void set_true(int atom, int chosen) 
{
	int rpos;
	struct lp_rlist *j;
	
	Model[atom] = LP_L_TRUE;
				
	/* push this atom */
	Stack[Sp] = atom;
	
	AuxStack[Sp++] = chosen ? AUX_POS : AUX_EXP_POS;

	/* We have put some rules closer to firing */
	for(j=PosDepends[atom]; j!=NULL; j=j->next) {
		rpos = lp_rpos(Prog,j->rule);
		--PosBodySize[rpos];
		maybe_fire(rpos);
	}
	/* No rule with this atom in its negative body can fire */
	for(j=NegDepends[atom]; j!=NULL; j=j->next) {
		rpos = lp_rpos(Prog,j->rule);
		suppress(rpos,0);
	}
	/* If this atom is the head of only one rule, that rule must have fired */
	if(HeadNumber[atom] == 1) {
		for(j=HeadOf[atom]; j!=NULL; j=j->next) {
			rpos = lp_rpos(Prog,j->rule);
			if(Suppressed[rpos] == 0) {
				backchain_true(rpos);
			}
		}
	}
}


int trypos(struct lp_atom *atom, int *SeenPos, int *SeenNeg) 
{
	int oldsp;
	int j,a;
	int pos;

	oldsp = Sp;

	DO_LOG3(LP_LOG_EXTRA, "trypos: trying atom %d (%s) positively", lp_apos(Prog,atom), atom->name);
	lp_aq_enqueue(PosQueue, atom);
	expand(1);
	++TriedAtoms;

	pos = 0;
	for(j=oldsp; j<Sp; j++)
		if(NAnt[Stack[j]])
			++pos;
	
	DO_LOG2(LP_LOG_EXTRA, "trypos: got %d new atoms", pos);
	
	if(!Conflict) {
		/* If we have found a stable model, return immediately */
		if(full_model())
			return -1;

		DO_LOG(LP_LOG_EXTRA, "trypos: marking atoms as seen");
		/* Do not later try things that were expanded */
		for(j=oldsp; j<Sp; j++) {
			a = Stack[j];
			if(Model[a] == LP_L_TRUE)
				SeenPos[a] = pos;
			else if(Model[a] == LP_L_FALSE)
				SeenNeg[a] = pos;
		}

		/* And backtrack */
		DO_LOG3(LP_LOG_EXTRA, "trypos: backtracking to Sp=%d (from %d)", oldsp, Sp);
		for(--Sp; Sp>=oldsp; --Sp)
			backtrack_single(Stack[Sp], AuxStack[Sp]);
		Sp = oldsp;
		DO_LOG(LP_LOG_EXTRA, "trypos: finished backtracking");
	} else {
		/* Replace the thing that conflicted with its negation */
		Conflict = 0;

		DO_LOG3(LP_LOG_EXTRA, "trypos: backtracking to Sp=%d (from %d)", oldsp, Sp);
		for(--Sp; Sp>=oldsp; --Sp)
			backtrack_single(Stack[Sp], AuxStack[Sp]);
		Sp = oldsp;
		DO_LOG(LP_LOG_EXTRA, "trypos: finished backtracking");
		
		DO_LOG2(LP_LOG_EXTRA, "trypos: adding not(%s)", atom->name);
		lp_aq_enqueue(NegQueue, atom);
		expand(0);
		return -1;
	}

	return pos;
}


int tryneg(struct lp_atom *atom, int *SeenPos, int *SeenNeg) 
{
	int oldsp;
	int j,a;
	int neg;

	oldsp = Sp;

	DO_LOG3(LP_LOG_EXTRA, "tryneg: trying atom %d (%s) negatively", lp_apos(Prog,atom), atom->name);
	lp_aq_enqueue(NegQueue, atom);
	expand(1);
	++TriedAtoms;

	neg = 0;
	for(j=oldsp; j<Sp; j++)
		if(NAnt[Stack[j]])
			++neg;
	
	DO_LOG2(LP_LOG_EXTRA, "tryneg: got %d new atoms", neg);
	
	if(!Conflict) {
		/* If we have found a stable model, return immediately */
		if(full_model())
			return -1;

		DO_LOG(LP_LOG_EXTRA, "tryneg: marking atoms as seen");
		/* Do not later try things that were expanded */
		for(j=oldsp; j<Sp; j++) {
			a = Stack[j];
			if(Model[a] == LP_L_TRUE)
				SeenPos[a] = neg;
			else if(Model[a] == LP_L_FALSE)
				SeenNeg[a] = neg;
		}

		/* And backtrack */
		DO_LOG3(LP_LOG_EXTRA, "tryneg: backtracking to Sp=%d (from %d)", oldsp, Sp);
		for(--Sp; Sp>=oldsp; --Sp)
			backtrack_single(Stack[Sp], AuxStack[Sp]);
		Sp = oldsp;
		DO_LOG(LP_LOG_EXTRA, "tryneg: finished backtracking");
	} else {
		/* Replace the thing that conflicted with its negation */
		Conflict = 0;
		
		DO_LOG3(LP_LOG_EXTRA, "tryneg: backtracking to Sp=%d (from %d)", oldsp, Sp);
		for(--Sp; Sp>=oldsp; --Sp)
			backtrack_single(Stack[Sp], AuxStack[Sp]);
		Sp = oldsp;
		DO_LOG(LP_LOG_EXTRA, "tryneg: finished backtracking");
		
		DO_LOG2(LP_LOG_EXTRA, "tryneg: adding %s positively", atom->name);
		lp_aq_enqueue(PosQueue, atom);
		expand(0);
		return -1;
	}

	return neg;
}


void handle_waiting_messages(void)
{
	struct message m;

	begin_time(TIM_NET, "handle_waiting_messages");
	while(pvm_nrecv(-1, -1) > 0) {
		if(parse_message(&m))
			break;
		
		switch(m.type) {
		    case MSG_WAITING:
			DO_LOG2(LP_LOG_INFO, "handle_waiting_messages: got MSG_WAITING from %x", m.from);
			mq_enqueue(WaitingMachines,m.from);
			break;
		    case MSG_DIE:
			DO_LOG2(LP_LOG_INFO, "handle_waiting_messages: got MSG_DIE from %x", m.from);
			DO_LOG2(LP_LOG_ERROR, "dying at request of %x", m.from);
			die(0);
		    case MSG_DIED:
			if(m.from == pvm_parent()) {
				DO_LOG(LP_LOG_ERROR, "handle_waiting_messages: parent died---I should too");
				die(1);
			}
			break;
		    case MSG_STACK:
		    case MSG_STACK_UPD:
			free(m.data);
			/* FALL-THROUGH */
		    default:
			not_expecting(m.type);
		}
	}
	end_time(TIM_NET, "handle_waiting_messages");
}


void maybe_branch(void) 
{
	machine_t machine=-1;
	int tfval, bottom=0;
	int result;
	int saved;
	int sent=0;
	int cps, i;
	int matches;
	char pbuf[96];
	machine_t me = pvm_mytid();

	begin_time(TIM_MISC, "maybe_branch");
	
	while(!mq_empty(WaitingMachines)) {
		/* Look for AUX_POS or AUX_NEG (guessed) */
		matches = 0;
		bottom = -1;
		if(BalanceDepth == 0) {
			for(i=Sp-1; i>=0; --i) {
				if(AuxStack[i] == AUX_POS || AuxStack[i] == AUX_NEG) {
					bottom = i;
					break;
				}
			}
		} else {
			for(i=0; i<Sp; ++i) {
				if(AuxStack[i] == AUX_POS || AuxStack[i] == AUX_NEG) {
					bottom = i;
					if(++matches == BalanceDepth)
						break;
				}
			}
		}

		if(bottom == -1) {
			DO_LOG(LP_LOG_INFO, "maybe_branch: nothing to split on");
			break;
		}

		DO_LOG4(LP_LOG_INFO, "maybe_branch: trying to split on Sp[%d] = %d (%s)",
			bottom, Stack[bottom], Prog->atoms[Stack[bottom]].name);

		do {
			machine = mq_dequeue(WaitingMachines);
			if(machine == me)
				DO_LOG(LP_LOG_ERROR, "maybe_branch: ERROR: I was in my own WaitingMachines queue");
		} while(machine == me);

		saved = AuxStack[bottom];
		if(saved == AUX_POS)
			tfval = AUX_EXP_POS;
		else
			tfval = AUX_EXP_NEG;

		DO_LOG2(LP_LOG_DEBUG, "maybe_branch: calling send_stack to %x", machine);

		AuxStack[bottom] = !tfval;
		if((result = send_stack_CH(machine,bottom+1,Stack,AuxStack))) {
			/* Restore the guessedness */
			AuxStack[bottom] = saved;
			DO_LOG3(LP_LOG_ERROR, "maybe_branch: send_stack to %x failed (%d)", machine, result);
			continue;
		}
		++sent;
		AuxStack[bottom] = tfval;
		
		DO_LOG2(LP_LOG_INFO, "maybe_branch: send_stack to %x succeeded", machine);
		if((result = send_sentjob(pvm_parent(), machine)))
			DO_LOG(LP_LOG_ERROR, "maybe_branch: ERROR sending MSG_SENTJOB to master!!");
		
	}

	/* If there are machines left, and we're *sure* we didn't leave any choice points */
	if(!mq_empty(WaitingMachines) && (sent==0 || BalanceDepth==0))
		clear_waiting();

	for(i=0, cps=0; i<Sp; ++i)
		if(AuxStack[i] == AUX_POS || AuxStack[i] == AUX_NEG)
			++cps;
	sprintf(pbuf, "maybe_branch to %d machines, CPs now %d", sent, cps);
	end_time(TIM_MISC, "maybe_branch");
}


void die(int v) 
{
	if(Prog)
		stable_free();
	
	DO_LOG2(LP_LOG_ERROR, "dying with %d", v);
	
	if(logfile)
		fclose(logfile);
	
	pvm_exit();
	stop_timing();
	exit(v);
}


void dump_model(void) 
{
	int i;
	for(i=0; i < Prog->natoms; ++i)
		printf("Atom %d (%s) = %d\n", i, Prog->atoms[i].name, Model[i]);
}


int log_2(unsigned x) 
{
	int i;
	
	for(i=0; x!=0; ++i)
		x >>= 1;
	
	return i;
}

void log_stack(void)
{
	int i;
	if(Logging >= LP_LOG_DEBUG) {
		begin_time(TIM_LOG, "log");
		fprintf(logfile, "[%x] STACK:", pvm_mytid()); 
		for(i=0; i<Sp; ++i)
			fprintf(logfile, " %s%s", AuxLetters[AuxStack[i]], Prog->atoms[Stack[i]].name);
		
		fprintf(logfile, "\n");
	       
		fflush(logfile); 
		end_time(TIM_LOG, "log"); 
	} 
}


