/* timing.c
 * Neil Moore <neil@cs.uky.edu>
 * 22 May 2001
 *
 * Part of parstab.  See COPYRIGHT for copyright information.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include <lp_util.h>
#include <pvm3.h>
#include "timing.h"
#include "misc.h"

/* Should we do timing logs? */
int Timing=0;
/* Should they be undetailed? */
int TimeShort=0;
const char *TimingPrefix="pstim";

static int write_time(int typ, int begin, const char *extra);
static int begin_log(void);


struct timetyp_l {
	int type;
	struct timetyp_l *next;
};

static struct timetyp_l *timestack = 0;
static FILE *timelog = 0;
static struct timeval lasttime;

static struct timeval meastimes[TIM_MAX+1];

int write_time(int typ, int begin, const char *extra) 
{
	struct timeval curr;
	struct timeval diff;
	struct timezone junk;

	if(Timing == 0)
		return 0;
	
	if(timelog == NULL)
		return -1;
	gettimeofday(&curr, &junk);

	diff.tv_usec = curr.tv_usec - lasttime.tv_usec;
	diff.tv_sec = curr.tv_sec - lasttime.tv_sec;
	if(diff.tv_usec < 0) {
		diff.tv_usec += 1000000;
		diff.tv_sec -= 1;
	}
	if(TimeShort) {
		meastimes[typ].tv_usec += diff.tv_usec;
		meastimes[typ].tv_sec += diff.tv_sec;
		if(meastimes[typ].tv_usec > 1000000) {
			meastimes[typ].tv_usec -= 1000000;
			meastimes[typ].tv_sec += 1;
		}
		gettimeofday(&lasttime, &junk);
		return 0;
	}
	if(timestack) {  /* If timestack == NULL, we aren't currently doing any timing */
		if(fprintf(timelog, "%ld %ld %ld %ld %s %s %d\n", curr.tv_sec, curr.tv_usec,
			   diff.tv_sec, diff.tv_usec, begin ? "begin" : "end", extra, typ) <= 0)
			return -1;
		fflush(timelog);
	}

	/* compute the time again, because writing could have taken some time */
	gettimeofday(&lasttime, &junk);
	
	return 0;
}


int begin_log(void) 
{
	char fname[80];
	struct timezone junk;
	int i;
	
	if(Timing == 0)
		return 0;
	
	sprintf(fname, "%s-%x.log", TimingPrefix, pvm_mytid());
	timelog = fopen(fname, "w");
	if(timelog == NULL)
		return errno;

	gettimeofday(&lasttime, &junk);
	for(i=0; i<=TIM_MAX; ++i) {
		meastimes[i].tv_sec = 0;
		meastimes[i].tv_usec = 0;
	}
	return 0;
}


int begin_time(int what, const char *extra) 
{
	struct timetyp_l *n;
	struct timezone junk;
	int result;	

	if(Timing == 0)
		return 0;
	
	if(timelog == NULL)
		result = begin_log();

	if(timestack) {
		result = write_time(timestack->type, 1, extra);
	} else {
		gettimeofday(&lasttime, &junk);
		fprintf(timelog, "%ld %ld %ld %ld begin %s %d\n", lasttime.tv_sec, lasttime.tv_usec,
			0L, 0L, extra, what);
		fflush(timelog);
		result = 0;
	}
	
	
	n = lp_calloc(1, sizeof(struct timetyp_l));
	n->type = what;
	n->next = timestack;
	timestack = n;

	return result;
}


int end_time(int what, const char *extra) 
{
	int result;
	struct timetyp_l *o = timestack;
	
	if(Timing == 0)
		return 0;
	
	if(o == NULL) {
		DO_LOG(LP_LOG_ERROR, "end_time: trying to pop from an empty stack");
		return -1;
	} else if(o->type != what) {
		DO_LOG3(LP_LOG_ERROR, "end_time: trying to pop %d, but stack has %d", what, o->type);
		return -1;
	}

	result = write_time(what, 0, extra);
		
	timestack = o->next;
	free(o);

	return result;
}


int stop_timing(void) 
{
	struct timetyp_l *o, *i = timestack;
	
	if(Timing == 0)
		return 0;
	
	if(timelog == NULL)
		return -1;

	if(i)
		write_time(i->type, 0, "stop");
	
	while(i) {
		o = i;
		i = i->next;
		free(o);
	}
	if(TimeShort) {
		int j;
		for(j=0; j<=TIM_MAX; ++j) {
			fprintf(timelog, "%d: %ld %ld\n", j, meastimes[j].tv_sec, meastimes[j].tv_usec);
		}
	}	
	fclose(timelog);
	timelog = NULL;
	
	return 0;
}
